#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import os
import numpy as np
from natf.cell import Cell
from natf.radwaste_standard import RadwasteStandard
from natf.utils import data_to_line_1d, format_single_output

gamma_energy_upper_bounds = [0.01, 0.02, 0.05, 0.10, 0.20, 0.30, 0.40, 0.60,
        0.80, 1.00, 1.22, 1.44, 1.66, 2.00, 2.50, 3.00,
        4.00, 5.00, 6.50, 8.00,10.00,12.00, 14.00, 20.00]

class Part(object):
    '''class Part'''

    def __init__(self):
        self._name = ''  # name of the part
        self._dirname = None # directory for output files
        self._cell_ids = [] # list of cell ids
        self._subpart_ids = []
        self._part_cell_list = []  # a lis of Cell
        self._part_subpart_list = [] # list of sub parts
        self._mass_flow_rate = 0.0 # mass flot rate of this part
        self._node_part_count = [] # number of BLK modules merge into this node

    def __str__(self):
        p_str = f"Part name: {self.name}"
        p_str = f"{p_str}\npart_cell_list:"
        for c in part_cell_list:
            p_str = f"{p_str} {c.id}"
        return p_str
        
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise ValueError('name of the part must be string')
        self._name = value

    @property
    def dirname(self):
        return self._dirname

    @dirname.setter
    def dirname(self, value):
        if not isinstance(value, str):
            raise ValueError('dirname of the part must be string')
        self._dirname = value

    @property
    def cell_ids(self):
        return self._cell_ids

    @cell_ids.setter
    def cell_ids(self, value):
        if not isinstance(value, list):
            raise ValueError('cell_ids must be a list')
        for i in range(len(value)):
            if not isinstance(value[i], int):
                raise ValueError('cell_ids must be a list of int')
        self._cell_ids = value

    @property
    def subpart_ids(self):
        return self._subpart_ids

    @subpart_ids.setter
    def subpart_ids(self, value):
        if not isinstance(value, list):
            raise ValueError('subpart_ids must be a list')
        for i in range(len(value)):
            if not isinstance(value[i], str):
                raise ValueError('subpart_ids must be a list of string')
        self._subpart_ids = value


    @property
    def part_cell_list(self):
        return self._part_cell_list

    @part_cell_list.setter
    def part_cell_list(self, value):
        if not isinstance(value, list):
            raise ValueError('part_cell_list must be a list')
        for i in range(len(value)):
            if not isinstance(value[i], Cell):
                raise ValueError('part_cell_ist must be a list of Cell')
        self._part_cell_list = value

    @property
    def part_subpart_list(self):
        return self._part_cell_list

    @part_subpart_list.setter
    def part_subpart_list(self, value):
        if not isinstance(value, list):
            raise ValueError('part_subpart_list must be a list')
        for i in range(len(value)):
            if not isinstance(value[i], Part):
                raise ValueError('part_subpart_ist must be a list of Cell')
        self._part_subpart_list = value


    @property
    def mass_flow_rate(self):
        return self._mass_flow_rate

    @mass_flow_rate.setter
    def mass_flow_rate(self, value):
        if not isinstance(value, float):
            raise ValueError('mass flow rate of the part must be float')
        if value < 0.0:
            raise ValueError('mass flow rate must >= 0.0')
        self._mass_flow_rate = value

    @property
    def node_part_count(self):
        return self._node_part_count

    @node_part_count.setter
    def node_part_count(self, value):
        if not isinstance(value, list) and not isinstance(value, np.ndarray):
            raise ValueError('node_part_count must be a list')
        for i in range(len(value)):
            if not isinstance(value[i], int) and not isinstance(value[i], np.int64):
                raise ValueError('node_part_count must be a list of int')
        self._node_part_count = value

    @property
    def equal_cell(self):
        return self._equal_cell

    def init_equal_cell(self):
        """
        Calculate the equal_cell:
        - name (use part name)
        - vol (sum)
        - mass (sum) 
        - density (average)
        - neutron flux (vol average)
        """
        self._equal_cell = Cell()
        self._equal_cell.name = self.name
        # check part_cell_list
        if len(self.part_cell_list) >= 1:
            # merge basic info. vol, mass and density
            for c in self.part_cell_list:  # vol and mass
                self.equal_cell.vol += c.vol
                self.equal_cell.mass += c.mass
            self.equal_cell.density = self.equal_cell.mass / self.equal_cell.vol  # density
            # neutron flux
            ave_flux = [0.0] * len(self.part_cell_list[0].neutron_flux)
            for i, c in enumerate(self.part_cell_list):
                for j in range(len(ave_flux)):
                    ave_flux[j] += c.vol / self.equal_cell.vol * c.neutron_flux[j]
            self.equal_cell.neutron_flux = ave_flux

    def merge_cell(self, AIM):
        """merge_cell: a method of class Part
        funciton: merge the part_cell_list to make a equivalent cell
        data need to merge: basic: vol, mass, density
                            result data: nuclide, half_life, act, decay_heat, ...
                            contact_dose, ci, dpa, He_production, H_production"""
        # chech AIM
        if AIM not in ('CELL_ACT_POST', 'CELL_DPA_POST'):
            raise ValueError(
                'Part.merge_cell() can only be called when the AIM is CELL_ACT/DPA_POST')
        if AIM == 'CELL_ACT_POST':
            # treat nuclide and half_life
            for c in self._part_cell_list:
                for i, nuc in enumerate(c.nuclide):
                    if nuc in self._equal_cell.nuclide:
                        pass
                    if nuc not in self._equal_cell.nuclide:
                        self._equal_cell.nuclide.append(nuc)
                        self._equal_cell.half_life.append(c.half_life[i])
            # treat act, decay_heat, contact_dose and ci
            # first, resize the act, decay_heat, contact_dose and ci
            NUC = len(self._equal_cell.nuclide)
            INTV = self._part_cell_list[0].act.shape[0]
            self._equal_cell.act = np.resize(self._equal_cell.act, (INTV, NUC))
            self._equal_cell.total_alpha_act = np.resize(self._equal_cell.total_alpha_act, (INTV))
            self._equal_cell.decay_heat = np.resize(
                self._equal_cell.decay_heat, (INTV, NUC))
            self._equal_cell.contact_dose = np.resize(
                self._equal_cell.contact_dose, (INTV, NUC))
            self._equal_cell.ci = np.resize(self._equal_cell.ci, (INTV, NUC))
            # merge the data
            for c in self._part_cell_list:
                for i, nuc in enumerate(c.nuclide):
                    nid = self._equal_cell._nuclide.index(nuc)
                    for intv in range(INTV):
                        self._equal_cell.act[intv][nid] += c.act[intv][i] * (
                            c.mass / self._equal_cell.mass)  # unit: Bq/kg
                        self._equal_cell.decay_heat[intv][nid] += c.decay_heat[intv][i] * (
                            c.mass / self._equal_cell.mass)  # kW/kg
                        self._equal_cell.contact_dose[intv][nid] += c.contact_dose[intv][i] * (
                            c.mass / self._equal_cell.mass)  # unit: Sv/h
                        self._equal_cell.ci[intv][nid] += c.ci[intv][i] * \
                            (c.mass / self._equal_cell.mass)
                for intv in range(INTV):
                    self._equal_cell.total_alpha_act[intv] += c.total_alpha_act[intv] * (c.mass / self._equal_cell.mass)
        if AIM == 'CELL_DPA_POST':
            # dpa and gas production
            dpa, He_production, H_production = 0.0, 0.0, 0.0
            for c in self._part_cell_list:
                dpa += c.dpa * (c.mass / self._equal_cell.mass)
                He_production += c.He_production * \
                    (c.mass / self._equal_cell.mass)
                H_production += c.H_production * \
                    (c.mass / self._equal_cell.mass)
            self._equal_cell.dpa = dpa
            self._equal_cell.He_production = He_production
            self._equal_cell.H_production = H_production

    def part_act_analysis(self, AIM, rwss=[]):
        """
        Analysis the:
        - acitivty,
        - contact_dose,
        - decay_heat,
        - ci,
        - radwaste class (optional),
        of the Part, using the method the Cell.

        Parameters:
        -----------
        AIM: AIM, must be 'CELL_ACT_POST'
        rwss: list of RadwasteStandard, optional
            The standards used.
            Supported standards are: CHN2018, UK, USNRC, USNRC_FETTER.
        """

        # AIM check
        if AIM not in ['CELL_ACT_POST']:
            raise ValueError(
                'method act_analysis() of a Part should be called only when the AIM is CELL_ACT_POST')
        # data prepare check
        if self._equal_cell.act.shape[0] < 1:
            raise ValueError(
                'method act_alalysis() should be called after filling the data of act')
        # analysis the act info.
        # calculate the act_max_nuc, act_max_act and act_max_ratio
        self._equal_cell.analysis_act()
        # analysis the decay_heat
        self._equal_cell.analysis_decay_heat()
        # analysis the contact_dose
        self._equal_cell.analysis_contact_dose()
        # analysis the ci
        self._equal_cell.analysis_ci()
        # analysis radwaste classification
        for rws in rwss:
            self._equal_cell.analysis_radwaste(rws=rws)

    def generate_directory(self, WORK_DIR):
        """
        Generate directory name for output files.
        """
        self.dirname = os.path.join(WORK_DIR, self._name)
        if not os.path.isdir(self.dirname):
            os.mkdir(self.dirname)

    def generate_output_filename(self, item):
        """
        Create output filename for specific file.
        """
        filename = os.path.join(self.dirname, ''.join([self._name, '.', item]))
        return filename
    
    def generate_output_title_line(self, item):
        """
        Generate output file title line for each item.
        """
        title_line = None
        # input item check
        if item == 'basicinfo':
            title_line = data_to_line_1d(key=self._name, value='basic_information')
        if item == 'nuc':
            title_line = data_to_line_1d(key='Nuclide', value='Half_life(s)')
        if item in ['acts', 'act', 'dhv', 'dh', 'cd', 'ci']:
            title_line = data_to_line_1d(key='Cooling_time(s)', value=self._equal_cell.nuclide)
        if item in ['act_st_t']:
            value = ['total_specific_act(Bq/kg)', 'total_activity(Bq)']
            title_line = data_to_line_1d(key='Cooling_time(s)', value=value)
        if item in ['acts_max_ratio']:
            title_line = data_to_line_1d(key='Cooling_time(s)', value=self._equal_cell.act_max_contri_nuc)
        if item in ['acts_max']:
            title_line = data_to_line_1d(key='Cooling_time(s)', value=['Total'] + self._equal_cell.act_max_contri_nuc)
        if item in ['dh_vt_t']:
            value = ['total_decay_heat_vol(kW/m3)',
                     'total_decay_heat(kW)']
            title_line = data_to_line_1d(key='Cooling_time(s)', value=value)
        if item in ['dhv_max']:
            title_line = data_to_line_1d(key='Cooling_time(s)', value=['Total'] + self._equal_cell.decay_heat_max_contri_nuc)
        if item in ['dhv_max_ratio']:
            title_line = data_to_line_1d(key='Cooling_time(s)', value=self._equal_cell.decay_heat_max_contri_nuc)
        if item in ['cdt']:
            title_line = data_to_line_1d(key='Cooling_time(s)', value='total_contact_dose(Sv/hr)')
        if item in ['cd_max']:
            title_line = data_to_line_1d(key='Cooling_time(s)', value=['Total'] + self._equal_cell.contact_dose_max_contri_nuc)
        if item in ['cd_max_ratio']:
            title_line = data_to_line_1d(key='Cooling_time(s)', value=self._equal_cell.contact_dose_max_contri_nuc)
        if item in ['cit']:
            title_line = data_to_line_1d(key='Cooling_time(s)', value=['total_ci'])
        if item in ['ci_max']:
            title_line = data_to_line_1d(key='Cooling_time(s)', value=self._equal_cell.ci_max_contri_nuc)
        if item in ['ci_max_ratio']:
            title_line = data_to_line_1d(key='Cooling_time(s)', value=['Total'] + self._equal_cell.ci_max_contri_nuc)
        if item in ['gamma_emit_rate']:
            title_line = data_to_line_1d(key='Energy upper boundary (MeV)', value=['Gamma emit rate (g/cc/s)'])
        if title_line is None:  # check
            raise ValueError("item {0} is not supported".format(item))
        return title_line

    def output_item(self, item, key, value):
        """
        Output item.

        Parameters:
        -----------
        item: string
            The postfix of the file.
        key: string, list or numpy array
            The key first column of the of the data.
        value: float, list or numpy array
            The data of the content to print.
        """
        filename = self.generate_output_filename(item)
        fo = open(filename, 'w')
        title_line = self.generate_output_title_line(item)
        fo.write(title_line)
        if isinstance(key, list) or isinstance(key, np.ndarray):
            for i in range(len(key)):
                line = data_to_line_1d(key=key[i], value=value[i])
                fo.write(line)
        else:
            line = data_to_line_1d(key=key, value=value)
            fo.write(line)
        fo.close()

    def output_basicinfo(self, MODEL_DEGREE, cooling_times_cul):
        """
        Output basic infomation.
        """
        item = 'basicinfo'
        key = ["volume(cm3)", "mass(g)", "density(g/cm3)"]
        value = [self._equal_cell.vol * 360.0 / MODEL_DEGREE,
                 self._equal_cell.mass * 360.0 / MODEL_DEGREE,
                 self._equal_cell.density]
        self.output_item(item=item, key=key, value=value)

    def output_flux(self):
        """Output flux"""
        # write .flx file
        filename = self.generate_output_filename('flx')
        fo = open(filename, 'w')
        for i in range(len(self.equal_cell.neutron_flux) - 1):
            fo.write(f"{format_single_output(self.equal_cell.neutron_flux[i])}\n")
        fo.close()

    def output_nuc(self):
        # output nuclide & half_life
        item = 'nuc'
        key = self._equal_cell.nuclide
        value = self._equal_cell.half_life
        self.output_item(item=item, key=key, value=value)

    def output_acts(self, cooling_times_cul):
        # output specific activity info. unit Bq/kg
        item = 'acts'
        key = cooling_times_cul
        value = self._equal_cell.act
        self.output_item(item=item, key=key, value=value)

    def output_act(self, MODEL_DEGREE, cooling_times_cul):
        item = 'act'
        key = cooling_times_cul
        value = np.multiply(self._equal_cell.act, 360.0/MODEL_DEGREE*self._equal_cell.mass/1e3)
        self.output_item(item=item, key=key, value=value)

    def output_act_st_t(self, MODEL_DEGREE, cooling_times_cul):
        item = 'act_st_t'
        key = cooling_times_cul
        value = np.array([self._equal_cell.total_act,
                          np.multiply(np.array(self._equal_cell.total_act), self._equal_cell.mass / 1e3 * 360.0 /MODEL_DEGREE)
                          ])
        value = value.transpose().copy()
        self.output_item(item, key=key, value=value)

    def output_acts_max(self, cooling_times_cul):
        item = 'acts_max'
        key = cooling_times_cul
        value = np.concatenate((np.array([self._equal_cell.total_act]).transpose(),
            self._equal_cell.act_max_contri_act), axis=1)
        self.output_item(item=item, key=key, value=value)

    def output_acts_max_ratio(self, cooling_times_cul):
        item = 'acts_max_ratio'
        key = cooling_times_cul
        value = self._equal_cell.act_max_contri_ratio
        self.output_item(item=item, key=key, value=value)

    def output_dhv(self, cooling_times_cul, factor):
        item = 'dhv'
        key = cooling_times_cul
        value = np.multiply(self._equal_cell.decay_heat, factor)
        self.output_item(item=item, key=key, value=value)

    def output_dh(self, cooling_times_cul, factor):
        item = 'dh'
        key = cooling_times_cul
        value = np.multiply(self._equal_cell.decay_heat, factor)
        self.output_item(item=item, key=key, value=value)

    def output_dh_vt_t(self, cooling_times_cul, factor1, factor2):
        item = 'dh_vt_t'
        key = cooling_times_cul
        value = np.array([np.multiply(self._equal_cell.total_decay_heat, factor1),
                          np.multiply(np.array(self._equal_cell.total_decay_heat), factor2)
                          ])
        value = value.transpose().copy()
        self.output_item(item=item, key=key, value=value)

    def output_dhv_max(self, cooling_times_cul, factor):
        item = 'dhv_max'
        key = cooling_times_cul
        value = np.concatenate((np.array([self._equal_cell.total_decay_heat]).transpose(),
            self._equal_cell.decay_heat_max_contri_dh), axis=1)
        value = np.multiply(value, factor)
        self.output_item(item=item, key=key, value=value)

    def output_dhv_max_ratio(self, cooling_times_cul):
        item = 'dhv_max_ratio'
        key = cooling_times_cul
        value = self._equal_cell.decay_heat_max_contri_ratio
        self.output_item(item=item, key=key, value=value)

    def output_cdt(self, cooling_times_cul):
        item ='cdt'
        key = cooling_times_cul
        value = self._equal_cell.total_contact_dose
        self.output_item(item=item, key=key, value=value)

    def output_cd(self, cooling_times_cul):
        item = 'cd'
        key = cooling_times_cul
        value = self._equal_cell.contact_dose
        self.output_item(item=item, key=key, value=value)

    def output_cd_max(self, cooling_times_cul):
        item = 'cd_max'
        key = cooling_times_cul
        value = np.concatenate((np.array([self._equal_cell.total_contact_dose]).transpose(),
            self._equal_cell.contact_dose_max_contri_cd), axis=1)
        self.output_item(item=item, key=key, value=value)

    def output_cd_max_ratio(self, cooling_times_cul):
        item = 'cd_max_ratio'
        key = cooling_times_cul
        value = self._equal_cell.contact_dose_max_contri_ratio
        self.output_item(item=item, key=key, value=value)

    def output_ci(self, cooling_times_cul):
        item = 'ci'
        key = cooling_times_cul
        value = self._equal_cell.ci
        self.output_item(item=item, key=key, value=value)
        
    def output_cit(self, cooling_times_cul):
        item = 'cit'
        key = cooling_times_cul
        value = np.array([self._equal_cell.total_ci])
        value = value.transpose().copy()
        self.output_item(item=item, key=key, value=value)

    def output_ci_max(self, cooling_times_cul):
        item = 'ci_max'
        key = cooling_times_cul
        value = np.concatenate((np.array([self._equal_cell.total_ci]).transpose(),
            self._equal_cell.ci_max_contri_ci), axis=1)
        self.output_item(item=item, key=key, value=value)

    def output_ci_max_ratio(self, cooling_times_cul):
        item = 'ci_max_ratio'
        key = cooling_times_cul
        value = self._equal_cell.ci_max_contri_ratio
        self.output_item(item=item, key=key, value=value)

    def output_gamma_emit_rate(self, gamma_energy_upper_bounds):
        item = 'gamma_emit_rate'
        key = gamma_energy_upper_bounds
        value = self.equal_cell.gamma_emit_rate[0]
        self.output_item(item=item, key=key, value=value)

    def output_data(self, WORK_DIR, MODEL_DEGREE, AIM, cooling_times_cul=None,
            rwss=[]):
        """
        Outputdata: output corresponding information according to the AIM and MODEL_DEGREE
        case 1: AIM = CELL_ACT_POST
            - output basic info. :vol, mass, density
            - nuclide info. : nuclide, half_life
            - act info. :nuclide, act
            - act_max_contri_nuc & act_max_contri_act
            - act_max_contri_nuc & act_max_contri_ratio
            - nuclide & decay_heat
            - decay_heat_max_contri_nuc & decay_heat_max_contri_dh
            - decay_heat_max_contri_nuc & decay_heat_max_contri_ratio
            - total_decay_heat
            - nuclide & contact_dose
            - contact_dose_max_contri_nuc & contact_dose_max_contri_cd
            - contact_dose_max_contri_nuc & contact_dose_max_contri_ratio
            - nuclide & ci
            - ci_max_contri_nuc & ci_max_contri_ci
            - ci_max_contri_nuc & ci_max_contri_ratio
            - radwaste classifcation, optional if rws is not None
        case 2: AIM = CELL_DPA_POST.
            - output basic info. : vol, mass, density
            - dpa,
            - He_production
            - H_production
        """

        # generate directory for this part
        self.generate_directory(WORK_DIR)
        # case 1: AIM == CELL_ACT_POST
        if AIM == 'CELL_ACT_POST':
            # check cooling_times_cul
            if cooling_times_cul is None:
                raise ValueError("cooling_times_cul must be provided in CELL_ACT_POST mode")
            # output basic info.
            self.output_basicinfo(MODEL_DEGREE, cooling_times_cul)
            # output nuclide & half_life
            self.output_nuc()
            # output flux
            self.output_flux()

            # -------------------activity part--------------------------------
            # output specific activity info. unit Bq/kg
            self.output_acts(cooling_times_cul)
            # output activity info. unit Bq
            self.output_act(MODEL_DEGREE, cooling_times_cul)
            # total_specific activy and total activity
            self.output_act_st_t(MODEL_DEGREE, cooling_times_cul)
            # output act_max_contri_nuc & act_max_contri_act, unit Bq/kg
            self.output_acts_max(cooling_times_cul)
            # output act_max_contri_nuc & act_max_contri_ratio, unit 1
            self.output_acts_max_ratio(cooling_times_cul)

            # --------------------decay_heat----------------------------------
            # output decay_heat info., unit kW/m3, origin kW/kg
            # unit change factor from kW/kg -> kW/m3
            factor = self._equal_cell.mass / 1.0e3 / (self._equal_cell.vol / 1.0e6)
            def f(x):
                return x * factor
            # decay heat kW/m3
            self.output_dhv(cooling_times_cul, factor)
            # output decay_heat of 360 degree model info. unit kW
            self.output_dh(cooling_times_cul, factor*self._equal_cell.vol/1e6)
            # total hecay_heat
            self.output_dh_vt_t(cooling_times_cul, factor, factor*self._equal_cell.vol/1e6)
            # output decay_heat_max_contri_nuc & decay_heat_max_contri_dh,
            # unit : kW/m3
            self.output_dhv_max(cooling_times_cul, factor)
            # output decay_heat_max_contri_nuc & decay_heat_max_contri_ratio,
            # unit 1
            self.output_dhv_max_ratio(cooling_times_cul)

            # ---------------------contact_dose-------------------------------
            # total contact_dose
            self.output_cdt(cooling_times_cul)
            # contact_dose of each nuc
            self.output_cd(cooling_times_cul)
            # contact_dose_max
            self.output_cd_max(cooling_times_cul)
            # contact_dose_max_ratio
            self.output_cd_max_ratio(cooling_times_cul)

            # ------------------------ci part---------------------------------
            self.output_ci(cooling_times_cul)
            # total ci
            self.output_cit(cooling_times_cul)
            # ci_max
            self.output_ci_max(cooling_times_cul)
            # ci_max_ratio
            self.output_ci_max_ratio(cooling_times_cul)

            # --------------------Radwaste part-------------------------------
            #  radwaste classification chn2018
            for rws in rwss:
                # radwaste classification results
                if rws.standard == 'CHN2018':
                    filename = self.generate_output_filename('rwc_chn2018')
                    line = ','.join(['Cooling_time(s)', 'Radwaste_Class', 'Decay_heat(kW/m3)',
                        'Clearance', 'VLLW', 'LLW', 'ILW', 'Limit'])
                    fo = open(filename, 'w')
                    fo.write(line+'\n')
                    for intv in range(len(cooling_times_cul)):
                        line = ','.join([format_single_output(cooling_times_cul[intv]), # cooling time
                            self._equal_cell.radwaste_class_chn2018[intv], # rwc
                            format_single_output(self._equal_cell.total_decay_heat[intv] * factor), # decay heat, (kW/m3)
                            data_to_line_1d(key=None, value=self._equal_cell.rw_chn2018_index_sum[intv], postfix=''), # indices
                            format_single_output(1.0)])  # limit
                        fo.write(line+'\n')
                    fo.close
                elif rws.standard == 'USNRC':
                    filename = self.generate_output_filename('rwc_usnrc')
                    line = ','.join(['Cooling_time(s)', 'Radwaste_Class', 'LLWA_LL',
                        'LLWB_LL', 'LLWC_LL', 'LLWA_SL', 'LLWB_SL', 'LLWC_SL', 'Limit'])
                    fo = open(filename, 'w')
                    fo.write(line+'\n')
                    for intv in range(len(cooling_times_cul)):
                        line = ','.join([format_single_output(cooling_times_cul[intv]), # cooling time
                            self._equal_cell.radwaste_class_usnrc[intv], # rwc
                            data_to_line_1d(key=None, value=self._equal_cell.rw_usnrc_index_sum_ll[intv], postfix=''),
                            data_to_line_1d(key=None, value=self._equal_cell.rw_usnrc_index_sum_sl[intv], postfix=''),
                            format_single_output(1.0)]) # limit
                        fo.write(line+'\n')
                    fo.close
                elif rws.standard == 'USNRC_FETTER':
                    filename = self.generate_output_filename('rwc_usnrc_fetter')
                    line = ','.join(['Cooling_time(s)', 'Radwaste_Class', 'LLWA',
                        'LLWB', 'LLWC', 'Limit'])
                    fo = open(filename, 'w')
                    fo.write(line+'\n')
                    for intv in range(len(cooling_times_cul)):
                        line = ','.join([format_single_output(cooling_times_cul[intv]), # cooling time
                            self._equal_cell.radwaste_class_usnrc_fetter[intv], # rwc
                            data_to_line_1d(key=None, value=self._equal_cell.rw_usnrc_fetter_index_sum[intv], postfix=''),
                            format_single_output(1.0)]) # limit
                        fo.write(line+'\n')
                    fo.close
                elif rws.standard == 'UK':
                    filename = self.generate_output_filename('rwc_uk')
                    line = ','.join(['Cooling_time(s)', 'Radwaste_Class', 'Decay_heat(kW/m3)',
                        'Alpha_act(Bq/kg)', 'Act(Bq/kg)', 'Limit'])
                    fo = open(filename, 'w')
                    fo.write(line+'\n')
                    for intv in range(len(cooling_times_cul)):
                        line = ','.join([format_single_output(cooling_times_cul[intv]), # cooling time
                            self._equal_cell.radwaste_class_uk[intv], # rwc
                            format_single_output(self._equal_cell.total_decay_heat[intv] * factor), # decay heat, (kW/m3)
                            format_single_output(self._equal_cell._total_alpha_act[intv]), # alpha act
                            format_single_output(self._equal_cell._total_act[intv]), # total act
                            format_single_output(1.0)])  # limit
                        fo.write(line+'\n')
                    fo.close
                elif rws.standard == 'RUSSIAN':
                    filename = self.generate_output_filename('rwc_russian')
                    line = ','.join(['Cooling_time(s)', 'Radwaste_Class', 'Decay_heat(kW/m3)',
                        'LLW', 'Limit'])
                    fo = open(filename, 'w')
                    fo.write(line+'\n')
                    for intv in range(len(cooling_times_cul)):
                        line = ','.join([format_single_output(cooling_times_cul[intv]), # cooling time
                            self._equal_cell.radwaste_class_russian[intv], # rwc
                            format_single_output(self._equal_cell.total_decay_heat[intv] * factor), # decay heat, (kW/m3)
                            data_to_line_1d(key=None, value=self._equal_cell.rw_russian_index_sum[intv], postfix=''), # indices
                            format_single_output(1.0)])  # limit
                        fo.write(line+'\n')
                    fo.close


        # case 2: AIM == CELL_DPA_POST
        # output the dpa information
        if AIM == 'CELL_DPA_POST':
            # output the DPA information
            filename = self.generate_output_filename('dpa')
            fo = open(filename, 'w')
            line = ' '.join(['DPA:', str(self._equal_cell.dpa), 'DPA/FPY\n'])
            fo.write(line)
            line = ' '.join(['He_production:', str(
                self._equal_cell.He_production), 'appm/FPY\n'])
            fo.write(line)
            line = ' '.join(['H_production:', str(
                self._equal_cell.H_production), 'appm/FPY\n'])
            fo.write(line)
            fo.close()

        # case 3: AIM == COOLANT_ACT_POST
        if AIM == 'COOLANT_ACT_POST':
            # output the act, decay_heat, contact_dose
            filename = self.generate_output_filename('coolant_response')
            fo = open(filename, 'w')
            line = data_to_line_1d(key='Nuclide', value=self.equal_cell.nuclide)
            fo.write(line)
            line = data_to_line_1d(key='Specific act (Bq/kg)', value=self.equal_cell.act[0])
            fo.write(line)
            line = data_to_line_1d(key='Decay heat (kW/kg)', value=self.equal_cell.decay_heat[0])
            fo.write(line)
            line = data_to_line_1d(key='Contact dose (Sv/h)', value=self.equal_cell.contact_dose[0])
            fo.write(line)
            fo.close()
            # output gamma_emit_rate
            self.output_gamma_emit_rate(gamma_energy_upper_bounds)
            # output neutron_emit_rate
            filename = self.generate_output_filename('neutron_emit_rate')
            fo = open(filename, 'w')
            line = ','.join(['Energy (MeV)', 'Nuetron emit rate (n/kg/s)'])
            fo.write(line+'\n')
            nid = self.equal_cell.nuclide.index('N17')
            line = ','.join(['0.848', format_single_output(self.equal_cell.act[0][nid]*0.95)])
            fo.write(line+'\n')
            fo.close()


def get_part_index(Parts, p):
    """get the index of part p in the Parts"""
    part_index = -1
    for i in range(len(Parts)):
        if p.name == Parts[i].name:  # found it
            part_index = i
    # check value
    if part_index == -1:
        raise ValueError(f'part {p.name} not found in the Parts')
    return part_index


def get_part_index_by_name(Parts, name):
    """get the index of part p in the Parts"""
    part_index = -1
    for i in range(len(Parts)):
        if name == Parts[i].name:  # found it
            part_index = i
    # check value
    if part_index == -1:
        raise ValueError(f'part {name} not found in the Parts')
    return part_index


def get_part_index_by_cell_id(parts, cid, find_last=False):
    """get the index of part p in parts according to specific cid"""
    pidx = -1
    for i, p in enumerate(parts):
        if cid in p.cell_ids:
            if find_last:
                pidx = i
            else:
                return i
    if pidx == -1:
        raise ValueError(f"cell: {cid} not in parts")
    return pidx

def is_cell_id_in_parts(parts, cid):
    """check whether a cell id is in parts"""
    for i, p in enumerate(parts):
        if cid in p.cell_ids:
            return True
    return False





import os
from warnings import warn

__version__ = '1.6.5'

if os.name == 'nt':
    p = os.environ['PATH'].split(';')
    lib = os.path.join(os.path.split(__file__)[0], 'lib')
    os.environ['PATH'] = ";".join([lib] + p)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

import numpy as np
from natf.material import Material


class Mesh(object):
    """class Mesh"""

    def __init__(self):
        self._geom = ''  # xyz/cyl
        self._id = -1
        self._xmin = -1.0
        self._xmax = -1.0
        self._ymin = -1.0
        self._ymax = -1.0
        self._zmin = -1.0
        self._zmax = -1.0
        self._rmin = -1.0
        self._rmax = -1.0
        self._tmin = -1.0
        self._tmax = -1.0
        self._origin = [0.0, 0.0, 0.0]  # used only when geom is cyl
        self._axs = [0, 0, 0]  # used only when geom is cyl
        self._vec = [0, 0, 0]  # used only when geom is cyl
        self._vol = -1.0
        self._mass = -1.0
        self._density = -1.0
        self._mesh_cell_list = []  # cell_ids, not Cells
        self._mesh_cell_counts = []
        self._mesh_cell_vol_fraction = []  # float
        self._mesh_mat_list = []  # mat ids, not Materials
        self._mesh_mat_counts = []
        self._mesh_mat_vol_fraction = []
        self._packing_factor = -1.0  # float
        # gammit emit rate of mesh, a 2D array
        self._gamma_emit_rate = np.zeros((0, 0), dtype=float, order='C')
        self._neutron_flux = []  # list of float, size 176 only
        self._neutron_flux_error = []  # list of float, size 176 only
        self._mat = Material()  # material of the mesh

    @property
    def geom(self):
        return self._geom

    @geom.setter
    def geom(self, value):
        if not isinstance(value, str):
            raise ValueError('geom of mesh must be string')
        if value != 'xyz' and value != 'cyl':
            raise ValueError('geom must be xyz or cyl. (in lower case)')
        self._geom = value

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if not isinstance(value, int):
            raise ValueError('id must be integer')
        if value < 0:
            raise ValueError('id must larger than 0')
        self._id = value

    @property
    def xmin(self):
        return self._xmin

    @xmin.setter
    def xmin(self, value):
        if not isinstance(value, float):
            raise ValueError('xmin of mesh must be float')
        self._xmin = value

    @property
    def xmax(self):
        return self._xmax

    @xmax.setter
    def xmax(self, value):
        if not isinstance(value, float):
            raise ValueError('xmax of mesh must be float')
        self._xmax = value

    @property
    def ymin(self):
        return self._ymin

    @ymin.setter
    def ymin(self, value):
        if not isinstance(value, float):
            raise ValueError('ymin of mesh must be float')
        self._ymin = value

    @property
    def ymax(self):
        return self._ymax

    @ymax.setter
    def ymax(self, value):
        if not isinstance(value, float):
            raise ValueError('ymax of mesh must be float')
        self._ymax = value

    @property
    def zmin(self):
        return self._zmin

    @zmin.setter
    def zmin(self, value):
        if not isinstance(value, float):
            raise ValueError('zmin of mesh must be float')
        self._zmin = value

    @property
    def zmax(self):
        return self._zmax

    @zmax.setter
    def zmax(self, value):
        if not isinstance(value, float):
            raise ValueError('zmax of mesh must be float')
        self._zmax = value

    @property
    def rmin(self):
        return self._rmin

    @rmin.setter
    def rmin(self, value):
        if not isinstance(value, float):
            raise ValueError('rmin of mesh must be float')
        self._rmin = value

    @property
    def rmax(self):
        return self._rmax

    @rmax.setter
    def rmax(self, value):
        if not isinstance(value, float):
            raise ValueError('rmax of mesh must be float')
        self._rmax = value

    @property
    def tmin(self):
        return self._tmin

    @tmin.setter
    def tmin(self, value):
        if not isinstance(value, float):
            raise ValueError('tmin of mesh must be float')
        self._tmin = value

    @property
    def tmax(self):
        return self._tmax

    @tmax.setter
    def tmax(self, value):
        if not isinstance(value, float):
            raise ValueError('tmax of mesh must be float')
        self._tmax = value

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        if not isinstance(value, list):
            raise ValueError('origin must be a list')
        if len(value) != 3:
            raise ValueError('origin must have a length of 3')
        for item in value:
            if not isinstance(item, float):
                raise ValueError('origin must be a list of float')
        self._origin = value

    @property
    def axs(self):
        return self._axs

    @axs.setter
    def axs(self, value):
        if not isinstance(value, list):
            raise ValueError('axs must be a list')
        if len(value) != 3:
            raise ValueError('axs must have a length of 3')
        for item in value:
            if not isinstance(item, int):
                raise ValueError('axs must be a list of int')
        self._axs = value

    @property
    def vec(self):
        return self._vec

    @vec.setter
    def vec(self, value):
        if not isinstance(value, list):
            raise ValueError('vec must be a list')
        if len(value) != 3:
            raise ValueError('vec must have a length of 3')
        for item in value:
            if not isinstance(item, int):
                raise ValueError('vec must be a list of int')
        self._vec = value

    @property
    def vol(self):
        return self._vol

    @vol.setter
    def vol(self, value):
        if not isinstance(value, float):
            raise ValueError('vol of mesh must be float')
        if value < 0.0:
            raise ValueError('vol must not smaller than 0.0')
        self._vol = value

    @property
    def mass(self):
        return self._mass

    @mass.setter
    def mass(self, value):
        if not isinstance(value, float):
            raise ValueError('mass of mesh must be float')
        if value < 0.0:
            raise ValueError('mass must not smaller than 0.0')
        self._mass = value

    @property
    def density(self):
        return self._density

    @density.setter
    def density(self, value):
        if not isinstance(value, float):
            raise ValueError('density of mesh must be float')
        if value < 0.0:
            raise ValueError('density must not smaller than 0.0')
        self._density = value

    @property
    def mesh_cell_list(self):
        return self._mesh_cell_list

    @mesh_cell_list.setter
    def mesh_cell_list(self, value):
        if not isinstance(value, list):
            raise ValueError('mesh_cell_list must a list')
        for i in range(len(value)):
            if not isinstance(value[i], int):
                raise ValueError('mesh_cell_list of mesh must be int')
            if value[i] < 1 or value[i] > 100000:
                raise ValueError(
                    'mesh_cell_list of mesh shoule between 1 ~ 100000')
        self._mesh_cell_list = value

    @property
    def mesh_cell_counts(self):
        return self._mesh_cell_counts

    @mesh_cell_counts.setter
    def mesh_cell_counts(self, value):
        if not isinstance(value, list):
            raise ValueError('mesh_cell_counts must be a list')
        for item in value:
            if not isinstance(item, int):
                raise ValueError('mesh_cell_counts must be a list of int')
            if item < 0:
                raise ValueError('mesh_cell_counts must no smaller than 0')
        self._mesh_cell_counts = value

    @property
    def mesh_cell_vol_fraction(self):
        return self._mesh_cell_vol_fraction

    @mesh_cell_vol_fraction.setter
    def mesh_cell_vol_fraction(self, value):
        if not isinstance(value, list):
            raise ValueError('mesh_cell_vol_fraction must a list')
        for i in range(len(value)):
            if not isinstance(value[i], float):
                raise ValueError(
                    'mesh_cell_vol_fraction of mesh must be float')
            if value[i] < 0.0 or value[i] > 1.0:
                raise ValueError(
                    'mesh_cell_vol_fraction of mesh shoule between 0 ~ 1')
        if abs(sum(value) - 1) > 1e-6:
            raise ValueError('sum of mesh_cell_vol_fraction must be 1.0')
        self._mesh_cell_vol_fraction = value

    @property
    def mesh_mat_list(self):
        return self._mesh_mat_list

    @mesh_mat_list.setter
    def mesh_mat_list(self, value):
        if not isinstance(value, list):
            raise ValueError('mesh_mat_list must a list')
        for i in range(len(value)):
            if not isinstance(value[i], int):
                raise ValueError('mesh_mat_list of mesh must be int')
            if value[i] < 0 or value[i] > 100000:
                raise ValueError(
                    'mesh_mat_list of mesh shoule between 0 ~ 100000')
        self._mesh_mat_list = value

    @property
    def mesh_mat_counts(self):
        return self._mesh_mat_counts

    @mesh_mat_counts.setter
    def mesh_mat_counts(self, value):
        if not isinstance(value, list):
            raise ValueError('mesh_mat_counts must be a list')
        for item in value:
            if not isinstance(item, int):
                raise ValueError('mesh_mat_counts must be a list of int')
            if item < 0:
                raise ValueError('mesh_mat_counts must no smaller than 0')
        self._mesh_mat_counts = value

    @property
    def mesh_mat_vol_fraction(self):
        return self._mesh_mat_vol_fraction

    @mesh_mat_vol_fraction.setter
    def mesh_mat_vol_fraction(self, value):
        if not isinstance(value, list):
            raise ValueError('mesh_mat_vol_fraction must a list')
        for i in range(len(value)):
            if not isinstance(value[i], float):
                raise ValueError('mesh_mat_vol_fraction of mesh must be float')
            if value[i] < 0.0 or value[i] > 1.0:
                raise ValueError(
                    'mesh_mat_vol_fraction of mesh shoule between 0 ~ 1')
        if abs(sum(value) - 1) > 1e-6:
            raise ValueError('sum of mesh_mat_vol_fraction must be 1.0')
        self._mesh_mat_vol_fraction = value

    @property
    def packing_factor(self):
        return self._packing_factor

    @packing_factor.setter
    def packing_factor(self, value):
        if not isinstance(value, float):
            raise ValueError('packing_factor of mesh must be float')
        if value < 0.0:
            raise ValueError('packing_factor must not smaller than 0.0')
        self._packing_factor = value

    @property
    def neutron_flux(self):
        return self._neutron_flux

    @neutron_flux.setter
    def neutron_flux(self, value):
        if not isinstance(value, list):
            raise ValueError('nuetron_flux of cell must be a list')
        if len(value) != 176:
            raise ValueError(
                'neutron flux should have data of 176, with the last of total data')
        for i in range(len(value)):
            if not isinstance(value[i], float):
                raise ValueError(
                    'neutron_flux of mesh must be a list of float')
            if value[i] < 0:
                raise ValueError(
                    'neutron_flux of mesh must be a list that each element is no smaller than 0.0')
        self._neutron_flux = value

    @property
    def neutron_flux_error(self):
        return self._neutron_flux_error

    @neutron_flux_error.setter
    def neutron_flux_error(self, value):
        if not isinstance(value, list):
            raise ValueError('neutron_flux_error must be a list')
        if len(value) != 176:
            raise ValueError(
                'neutron flux error of mesh should have data of 176, with the last of total data')
        for i in range(len(value)):
            if not isinstance(value[i], float):
                raise ValueError('neutron_flux_error must be a list of float')
            if value[i] < 0.0 or value[i] > 1:
                raise ValueError('neutron_flux_error must between 0 and 1')
        self._neutron_flux_error = value

    @property
    def gamma_emit_rate(self):
        return self._gamma_emit_rate

    @gamma_emit_rate.setter
    def gamma_emit_rate(self, value):
        if not isinstance(value, type(np.ndarray((0, 0)))):
            raise ValueError('gamma_emit_rate must be ndarray')
        if len(value.shape) != 2:
            raise ValueError(
                'gamma_emit_rate must be a two-dimensional ndarray')
        for i in range(value.shape[0]):
            for j in range(value.shape[1]):
                if not isinstance(value[i, j], float):
                    raise ValueError('gamma_emit_rate must composed of float')
        if value.min() < 0.0:
            raise ValueError('gamma_emit_rate must no smaller than 0.0')
        self._gamma_emit_rate = value

    @property
    def mat(self):
        return self._mat

    @mat.setter
    def mat(self, value):
        if not isinstance(value, Material):
            raise ValueError('mesh mat must be a "Material"')
        self._mat = value

    def cal_vol(self):
        if self._geom == 'xyz':
            self._vol = (self._xmax - self._xmin) * \
                (self._ymax - self._ymin) * (self._zmax - self._zmin)
        if self._geom == 'cyl':
            self._vol = 3.1415926 * (self._rmax ** 2 - self._rmin ** 2) * \
                (self._zmax - self._zmin) * (self._tmax - self._tmin)
        if self._vol < 0.0:
            raise ValueError(
                'there are mesh with vol smaller then 0.0, check the geometry')



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
import numpy as np
from natf.utils import mcnp_style_str_append

# mcnp mnemoics from user's guide vol II, chap III, p. 3-13
mnemoics = {'P':4, 'PX':1, 'PY':1, 'PZ':1,
            'SO':1, 'S':4, 'SX':2, 'SY':2, 'SZ':2,
            'C/X':3, 'C/Y':3, 'C/Z':3, 'CX':2, 'CY':2, 'CZ':2,
            'K/X':5, 'K/Y':5, 'K/Z':5, 'KX':3, 'KY':3, 'KZ':3,
            'SQ':10, 'GQ':10, 'TX':6, 'TY':6, 'TZ':6} # XYZP is not supported

class Surface(object):
    ''' class Surface'''

    def __init__(self, name=None, id=None, flag=None, mnemoic=None, card_entries=None):
        if name is not None:
            self._name = name
        if id is not None:
            self._id = id # id
        if flag is not None:
            self._flag = flag # * or + for boundaries
        if mnemoic is not None:
            self._mnemoic = mnemoic # type
        if card_entries is not None:
            self._card_entries = card_entries # parameters

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise ValueError('name must be string')
        self._name = value

    # id setter and getter
    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if not isinstance(value, int):
            raise ValueError('surface id must be integer')
        if value < 1 or value > 100000:
            raise ValueError('surface id must between 1 and 100000')
        self._id = value

    @property
    def flag(self):
        return self._flag

    @flag.setter
    def flag(self, value):
        if not isinstance(value, str):
            raise ValueError('surface flag must be string')
        if value not in ('*', '+'):
            raise ValueError("surface flag does not support!",
                "Supported flags: '*', '+'")
        self._flag = value

    @property
    def mnemoic(self):
        return self._mnemoic
    @mnemoic.setter
    def mnemoic(self, value):
        if not isinstance(value, str):
            raise ValueError('surface mnemoic must be string')
        if value.upper() not in mnemoics.keys():
            raise ValueError('surface mnemoic not supported')
        self._mnemoic = value.upper()

    @property
    def card_entries(self):
        return self._card_entries
    @card_entries.setter
    def card_entries(self, value):
        if not hasattr(self, 'mnemoic') or self.mnemoic is None:
            raise ValueError('mnemoic must be assigned before card entries')
        if not isinstance(value, list):
            raise ValueError('card_entries should be a list')
        if len(value) != mnemoics[self.mnemoic]:
            raise ValueError('wrong card entries number')
        self._card_entries = list(value)

    def __str__(self):
        """Return the mcnp style surface card"""
        s = ''.join([self.flag, str(self.id), '     ', self.mnemoic])
        indent_length = len(s)
        indent_str = ' '*indent_length
        for i, value in enumerate(self.card_entries):
            s = mcnp_style_str_append(s, value, indent_length)
        s = ''.join([s, '\n'])
        return s



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
import os
''' class Material.'''

thisdir = os.path.dirname(os.path.abspath(__file__))
appendix16 = os.path.join(thisdir, "data", "appendix16")

class Material(object):
    """calsee Material, used to store material information"""

    def __init__(self):
        self._id = -1
        self._atom_density = -1.0
        self._density = -1.0
        self._mcnp_material_nuclide = []
        self._mcnp_material_atom_fraction = []
        self._mcnp_material_mass_fraction = []
        self._fispact_material_nuclide = []
        self._fispact_material_atom_kilogram = []

    def __str__(self, card='mat', style='mcnp'):
        """Return difinition for material"""
        if card == 'mat' and style == 'mcnp':
            m_str = f"M{self._id} "
            if len(self.mcnp_material_atom_fraction) > 0:
                for i, nuc in enumerate(self.mcnp_material_nuclide):
                    m_str = f"{m_str}\n      {nuc} {self.mcnp_material_atom_fraction[i]}"
            elif len(self.mcnp_material_mass_fraction) > 0:
                for i, nuc in enumerate(self.mcnp_material_nuclide):
                    m_str = f"{m_str}\n      {nuc} -{self.mcnp_material_mass_fraction[i]}"
            else:
                raise ValueError(f"material {self.id} do not has nuclide composition")
            return m_str
        elif card == 'cell' and style=='mcnp':
            # create material string in cell card
            m_str = ''
            if self.density > 0:
                m_str = f"{self.id} -{self.density}"
            elif self.atom_density > 0:
                m_str = f"{self.id} {self.atom_density}"
            else:
                raise ValueError(f"material {self.id} does not assigned density/atom_density")
            return m_str
        else:
            raise ValueError(f"card: {card}, style: {style} not supported")
        
            

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if not isinstance(value, int):
            raise ValueError('id must be an integer!')
        if value > 100000 or value < 0:
            raise ValueError(
                'id must between 0 ~ 100000(the max cell number of MCNP')
        self._id = value

    @property
    def atom_density(self):
        return self._atom_density

    @atom_density.setter
    def atom_density(self, value):
        if not isinstance(value, float) and not isinstance(value, int):
            raise ValueError('atom_density must be a float')
        if value < 0:
            raise ValueError('atom_density should be larger than 0')
        if value > 1:
            raise ValueError(
                'atom_density is larger than 1.0, No material fits')
        self._atom_density = value

    @property
    def density(self):
        return self._density

    @density.setter
    def density(self, value):
        if not isinstance(value, float) and not isinstance(value, int):
            raise ValueError('density should be a float!')
        if value < 0:
            raise ValueError('density should be larger than 0!')
        if value > 30:
            raise ValueError(
                'density larger than 30, greater than the densitiest material in the world!')
        self._density = value

    @property
    def mcnp_material_nuclide(self):
        return self._mcnp_material_nuclide

    @mcnp_material_nuclide.setter
    def mcnp_material_nuclide(self, value):
        if not isinstance(value, list):
            raise ValueError('mcnp_material_nuclide should be a list')
        for i in range(len(value)):
            if not isinstance(value[i], str):
                raise ValueError(
                    'mcnp_material_nuclide should be a list of string')
            if not value[i].isdigit():
                raise ValueError(
                    'mcnp_material_nuclide should be a list of string that composed of digit')
            if len(value[i]) < 3 or len(value[i]) > 6:
                raise ValueError(
                    'mcnp_material_nuclide should be a list of string with length between 3~6')
        self._mcnp_material_nuclide = value

    @property
    def mcnp_material_atom_fraction(self):
        return self._mcnp_material_atom_fraction

    @mcnp_material_atom_fraction.setter
    def mcnp_material_atom_fraction(self, value):
        if not isinstance(value, list):
            raise ValueError('mcnp_material_atom_fraction should be a list')
        fraction_sum = 0.0
        for i in range(len(value)):
            if not isinstance(
                    value[i],
                    float) and not isinstance(
                    value[i],
                    int):
                raise ValueError(
                    'mcnp_material_atom_fraction should be a list of float/int')
            if value[i] < 0 or value[i] > 1:
                raise ValueError(
                    'mcnp_material_atom_fraction should be a list of float between 0 ~ 1')
            fraction_sum += value[i]
        if abs(fraction_sum - 1) > 1e-3:
            raise ValueError(
                'mcnp_material_atom_fraction should be a list that sum to be 1.0')
        if len(self._mcnp_material_nuclide) != 0 and len(
                value) != len(self._mcnp_material_nuclide):
            raise ValueError(
                'the lenght of mcnp_material_atom_fraction should equals to the length of mcnp_material_nuclide! check them')
        self._mcnp_material_atom_fraction = value

    @property
    def mcnp_material_mass_fraction(self):
        return self._mcnp_material_mass_fraction

    @mcnp_material_mass_fraction.setter
    def mcnp_material_mass_fraction(self, value):
        if not isinstance(value, list):
            raise ValueError('mcnp_material_mass_fraction should be a list')
        fraction_sum = 0.0
        for i in range(len(value)):
            if not isinstance(
                    value[i],
                    float) and not isinstance(
                    value[i],
                    int):
                raise ValueError(
                    'mcnp_material_mass_fraction should be a list of float/int')
            if value[i] < 0 or value[i] > 1:
                raise ValueError(
                    'mcnp_material_mass_fraction should be a list of float between 0 ~ 1')
            fraction_sum += value[i]
        if abs(fraction_sum - 1) > 1e-3:
            raise ValueError(
                'mcnp_material_mass_fraction should be a list that sum to be 1.0')
        if len(self._mcnp_material_nuclide) != 0 and len(
                value) != len(self._mcnp_material_nuclide):
            raise ValueError(
                'the lenght of mcnp_material_mass_fraction should equals to the length of mcnp_material_nuclide! check them')
        self._mcnp_material_mass_fraction = value

    # fispact_material_nuclide is a read only variable, so it doesn't have a
    # setter
    @property
    def fispact_material_nuclide(self):
        return self._fispact_material_nuclide

    # fispact_material_atom_kilogram is a read only variable, so it doesn't
    # have a setter
    @property
    def fispact_material_atom_kilogram(self):
        return self._fispact_material_atom_kilogram

    # convert the mcnp material to fispact material
    def mat_mcnp2fispact(self):
        if len(self._mcnp_material_nuclide) == 0:
            return
        if self._density < 0.0:
            raise Exception(
                'the density of the material should be set before convert')
        # convert the mcnp nuclide to fispact nuclide
        ELE_TABLE = ('H','HE','LI','BE','B',
                     'C','N','O','F','NE',
                     'NA','MG','AL','SI','P',
                     'S','CL','AR','K','CA',
                     'SC','TI','V','CR','MN',
                     'FE','CO','NI','CU','ZN',
                     'GA','GE','AS','SE','BR',
                     'KR','RB','SR','Y','ZR',
                     'NB','MO','TC','RU','RH',
                     'PD','AG','CD','IN','SN',
                     'SB','TE','I','XE','CS',
                     'BA','LA','CE','PR','ND',
                     'PM','SM','EU','GD','TB',
                     'DY','HO','ER','TM','YB',
                     'LU','HF','TA','W','RE',
                     'OS','IR','PT','AU','HG',
                     'TL','PB','BI','PO','AT',
                     'RN','FR','RA','AC','TH',
                     'PA','U','NP','PU','AM',
                     'CM','BK','CF','ES','FM')
        for i in range(len(self._mcnp_material_nuclide)):
            nuc = self._mcnp_material_nuclide[i]
            p = int(nuc) // 1000  # get the p number
            a = int(nuc) % 1000  # get the a number
            scale = self._atom_density * \
                self._mcnp_material_atom_fraction[i] * 1e24 * 1e3 / self._density
            if a != 0:
                self._fispact_material_nuclide.append(
                    ELE_TABLE[p - 1] + str(a))
                self._fispact_material_atom_kilogram.append(scale)
            if a == 0:
                # open the abundance file
                fin = open(appendix16)
                for line in fin:
                    line_ele = line.split()
                    if int(line_ele[0]) == p:
                        if (line_ele[1]) == '0.0':
                            errormessage = ''.join(
                                ['Nuclide: ', nuc, ' has no natural isotopes\n'])
                            raise ValueError(errormessage)
                        # get the start number
                        start_a = int(line_ele[3])
                        abundance_list = [float(x) for x in line_ele[4:]]
                        a_list = [x + start_a for x in range(len(abundance_list))]
                        isotopes = []
                        for item in a_list:
                            tempiso = ''.join([ELE_TABLE[p - 1], str(item)])
                            isotopes.append(tempiso)
                        # treat the isotope that has 0.0 abundance
                        real_isotopes = []
                        real_abundance_list = []
                        for i in range(len(abundance_list)):
                            # this is a isotope with 0.0 abundance
                            if abundance_list[i] < 1e-6:
                                pass
                            else:
                                real_abundance_list.append(abundance_list[i])
                                real_isotopes.append(isotopes[i])
                        self._fispact_material_nuclide.extend(real_isotopes)
                        self._fispact_material_atom_kilogram.extend(
                            map(lambda x: x * scale / 100.0, real_abundance_list))
                fin.close()

def create_pseudo_mat(mid=1, den=1.0):
    """
    Create a pseudo-material.
    """
    mat = Material()
    mat.id = mid
    mat.density = den
    mat.mcnp_material_nuclide = ['1001']
    mat.mcnp_material_mass_fraction = [1.0]
    return mat





#!/usr/bin/env python3
# -*- coding:utf-8 -*-
''' class Radwaste.'''
import sys
import os
import numpy as np
import pandas as pd
import re
from natf.utils import is_short_live, ci2bq

if sys.version_info[0] > 2:
    basestring = str

thisdir = os.path.dirname(os.path.abspath(__file__))

class RadwasteStandard(object):
    """Class Radwaste, used to store radioactive waste classification information"""

    def __init__(self, standard):
        self._standard = standard
        self._files = self.get_standard_default_files()
        self._classes = [] # names of the different classes
        self._limits = [] # limits of different classes
        self._df = None # CHN2018 use
        self._dfl = None # USNRC long live table use
        self._dfs = None # USNRC short live table use
        self.read_standard()

    @property
    def standard(self):
        return self._standard

    @standard.setter
    def standard(self, value):
        if not isinstance(value, str):
            raise ValueError('standard must be a string!')
        if value not in ('CHN2018', 'USNRC', 'USNRC_FETTER', 'UK'):
            raise ValueError("standard {0} not supported!".format(value))
        self._standard = value

    @property
    def files(self):
        return self._files

    @files.setter
    def files(self, value):
        if not isinstance(value, list):
            raise ValueError("files must be a list!")
        for f in value:
            if not os.path.isfile(f):
                raise ValueError("File {0} not found".format(f))
        self._flies = value

    @property
    def classes(self):
        return self._classes

    @classes.setter
    def classes(self, value):
        if not isinstance(value, list):
            raise ValueError("classes must be a list!")
        for item in value:
            if not isinstance(item, str):
                raise ValueError("class name should be string")
        self._classes = value

    @property
    def limits(self):
        return self._limits

    @limits.setter
    def limits(self, limits):
        if not isinstance(value, list):
            raise ValueError("limits must ba a list")
        for item in value:
            if not isinstance(item, map):
                raise ValueError("limit for each class must be a map")
        self._limits = value

    @property
    def df(self):
        return self._df

    @df.setter
    def df(self, value):
        if not isinstance(value, pd.DataFrame):
            raise ValueError("df must be a dataframe.")
        self._df = value

    @property
    def dfl(self):
        return self._dfl

    @dfl.setter
    def dfl(self, value):
        if not isinstance(value, pd.DataFrame):
            raise ValueError("dfl must be a dataframe.")
        self._dfl = value

    @property
    def dfs(self):
        return self._dfs

    @dfs.setter
    def dfs(self, value):
        if not isinstance(value, pd.DataFrame):
            raise ValueError("dfs must be a dataframe.")
        self._dfs = value

    def get_standard_default_files(self):
        """Get filename for specific standard."""
        files = []
        standard_dir = os.path.join(thisdir, "radwaste_standards", self.standard)
        if self.standard == 'CHN2018':
            files.append(os.path.join(standard_dir, self.standard + ".csv"))
        elif self.standard == 'USNRC':
            # long lvie table
            files.append(os.path.join(standard_dir, self._standard + "_LL.csv"))
            # short live table
            files.append(os.path.join(standard_dir, self._standard + "_SL.csv"))
        elif self.standard == 'USNRC_FETTER':
            files.append(os.path.join(standard_dir, self._standard + ".csv"))
        elif self.standard == 'RUSSIAN':
            files.append(os.path.join(standard_dir, self._standard + ".csv"))
        return files

    def read_standard(self):
        """Read the csv into dataframe."""
        if self.standard == 'CHN2018':
            df = pd.read_csv(self.files[0])
            self._df = df
            self._classes = df.columns[1:]
        elif self.standard == 'USNRC':
            dfl = pd.read_csv(self.files[0])
            dfs = pd.read_csv(self.files[1])
            self._dfl = dfl
            self._dfs = dfs
            self._classes = ['LLWA', 'LLWB', 'LLWC']
        elif self.standard == 'USNRC_FETTER':
            df = pd.read_csv(self.files[0])
            self._df = df
            self._classes = ['LLWA', 'LLWB', 'LLWC']
        elif self.standard == 'UK':
            self._classes = ['LLW', 'ILW', 'HLW']
        elif self.standard == 'RUSSIAN':
            self._classes = ['LLW', 'ILW']
            self._df = pd.read_csv(self.files[0])

    def get_nuc_limits_chn2018(self, nuc):
        """Get the limits for different class of a specific nuclide.
        
        Parameters:
        -----------
        nuc: string
            Nuclide name. Eg. H3, Cs137.

        Returns:
        --------
        Limit of the nuclide in specific material, unit [Bq/kg].
        """
        col_num = len(self.df.columns)
        limits = np.array(self.df.loc[self.df['Nuclide'] == nuc]).flatten()
        limits = limits[1:]
        # nuc not found in the list, set to default inf
        if len(limits) == 0:
            limits = np.array([float('inf')]*(col_num-1))
        return limits
        
    def get_nuc_limits_usnrc(self, nuc, half_life, density):
        """Get the limits for different class of a specific nuclide.
        
        Parameters:
        -----------
        nuc: string
            Nuclide name. Eg. H3, Cs137.
        half_life: float
            Half life of the nuclide, unit: [s].
        density: float
            Density of the material where the nuclide exists, unit: [g/cm3].
            Required in 'USNRC' and 'USNRC_FETTER' standard because of the
            unit conversion.

        Returns:
        --------
        Limit of the nuclide in specific material, unit [Bq/kg].
        """

        # check if the half life provided
        if half_life is None:
            raise ValueError("Half of the nuclide must be provided in "
            "USNRC standard")

        if nuc in ('Pu241', 'Cm242') or (not is_short_live(half_life)):
            # long lvie nuclide or Pu242/Cm242
            col_num = len(self.dfl.columns)
            # use data from USNRC_LL
            limits = np.array(self.dfl.loc[self.dfl['Nuclide'] == nuc]).flatten()
            limits = limits[1:]
        else:
            #  short-live 
            col_num = len(self.dfs.columns)
            # use data from USNRC_SL
            if half_life < 3600.0*24*365.25*5:
                limits = np.array([700.0, 700.0, 700.0])
            else:
                limits = np.array(self.dfs.loc[self.dfs['Nuclide'] == nuc]).flatten()
                limits = limits[1:]

        # convert the unit from Ci/m3 or Ci/g to Bq/kg
        if nuc in ('Pu241', 'Cm242'):
            # convert from Ci/g to Bq/kg
            limits = np.multiply(limits, ci2bq(1.0)*1e3)
        else:
            # convert from Ci/m3 to Bq/kg
            limits = np.multiply(limits, ci2bq(1.0)/density/1e3)

        # nuc not found in the list, set to default inf
        if len(limits) == 0:
            limits = np.array([float('inf')]*(col_num-1))
        return limits

    def get_nuc_limits_usnrc_fetter(self, nuc, half_life, density):
        """Get the limits for different class of a specific nuclide.
        
        Parameters:
        -----------
        nuc: string
            Nuclide name. Eg. H3, Cs137.
        half_life: float
            Half life of the nuclide, unit: [s].
        density: float
            Density of the material where the nuclide exists, unit: [g/cm3].
            Required in 'USNRC' and 'USNRC_FETTER' standard because of the
            unit conversion.

        Returns:
        --------
        Limit of the nuclide in specific material, unit [Bq/kg].
        """

        # check density
        if density is None:
            raise ValueError("Density must be provide in 'USNRC_FETTER'")
        if half_life < 3600.0*24*365.25*5:
            limits = np.array([700.0, 700.0, 700.0])
            # convert from Ci/m3 to Bq/kg
            limits = np.multiply(limits, ci2bq(1.0)/density/1e3)
        else:
            limits_s = np.array(self.df.loc[self.df['Nuclide'] == nuc]).flatten()
            if len(limits_s) > 1:
                limit_c = convert_nrc_fetter_limit(limits_s[-1], density=density)
                limits = [limit_c*0.01, limit_c*0.1, limit_c]
            else:
                # nuc not in table
                limits = [float('inf'), float('inf'), float('inf')]
        return limits

    def get_nuc_limits_russian(self, nuc, half_life=None):
        """Get the limits for different class of a specific nuclide.
        
        Parameters:
        -----------
        nuc: string
            Nuclide name. Eg. H3, Cs137.
        half_life: float
            Half life of the nuclide, unit: [s].

        Returns:
        --------
        Limit of the nuclide in specific material, unit [Bq/kg].
        """

        if half_life is not None:
            if is_short_live(half_life, threshold=5):
                limits = [float('inf')] 
                return limits
        else:
            # read table first
            limits = np.array(self.df.loc[self.df['Nuclide'] == nuc]).flatten()
            limits = limits[1:]
            if len(limits) == 0:
                # nuc not in table
                if is_u(nuc) or is_alpha_5(nuc):
                    limits = [3.7e6] 
                else:
                    limits = [float('inf')]
            else:
                return limits
        return limits

    def get_nuc_limits(self, nuc, half_life=None, density=None):
        """Get the limits for different class of a specific nuclide.
        
        Parameters:
        -----------
        nuc: string
            Nuclide name. Eg. H3, Cs137.
        half_life: float, optional
            Half life of the nuclide, unit: s. Required in USNRC standard.
        density: float, optional
            Density of the material where the nuclide exists, unit: g/cm3.
            Required in 'USNRC' and 'USNRC_FETTER' standard because of the
            unit conversion.

        Returns:
        --------
        Limit of the nuclide in specific material, unit Bq/kg.
        """

        # CHN2018 standard
        if self.standard == 'CHN2018':
            return self.get_nuc_limits_chn2018(nuc)
        # USNRC standard
        elif self.standard == 'USNRC':
            return self.get_nuc_limits_usnrc(nuc, half_life, density)
        # USNRC_FETTER standard
        elif self.standard == 'USNRC_FETTER':
            return self.get_nuc_limits_usnrc_fetter(nuc, half_life, density)
        # RUSSIAN standard
        elif self.standard == 'RUSSIAN':
            return self.get_nuc_limits_russian(nuc, half_life)

    def determin_class_chn2018(self, indices, decay_heat):
        """Determin the radwaste class CHN2018, according to the indices and
        the decay heat.
        
        Parameters:
        -----------
        indices: numpy array (1D)
            Indices of difference class of a specific cooling time.
            Eg. clear index, VLLW index, ...
        decay_heat: float
            Decay heat of a cooling time. Unit: kW/m3.
        """

        # if the decay heat is above 2kW/m3, it's HLW
        #if decay_heat > 2:
        #    return 'HLW'
        # if the decay heat is no more than 2kW/m3, check the indices
        # check indices length
        if len(indices) != len(self.df.columns) -1:
            raise ValueError("indices length wrong")
        for i in range(len(indices)):
            if indices[i] <= 1.0:
                return self.df.columns[1+i]
        # not any index no more than 1, higher class
        return 'ILW'

    def determin_class_usnrc(self, rwi_ll, rwi_sl):
        """Determin the radwaste class USNRC, according to the rw.
        
        Parameters:
        -----------
        rwi_ll: numpy array (1D)
            Indices of difference class of a specific cooling time for long
            lived nuclides and Pu241 and Cm242.
            Eg. LLWA, LLWB, LLWC, ILW
        rwi_sl: numpy array (1D)
            Indices of difference class of a specific cooling time for short
            lived nuclides.
        """

        classes = ['LLWA', 'LLWB', 'LLWC', 'ILW']
        # check indices length
        if len(rwi_ll) != len(self.dfl.columns) -1 or \
                (len(rwi_sl) != len(self.dfl.columns) -1):
            raise ValueError("radwaste indices length wrong")

        class_ll, class_sl, class_al = 3, 3, 3
        # class according to rwi_ll
        for i in range(len(rwi_ll)):
            if rwi_ll[i] <= 1.0:
                class_ll = i
                break
        # class according to rwi_sl
        for i in range(len(rwi_sl)):
            if rwi_sl[i] <= 1.0:
                class_sl = i
                break

        # combine two class
        if class_ll == 0:
            class_al = class_sl
        elif class_ll > 0 and class_ll <= 2:
            if class_sl <= 2:
                class_al = 2

        return classes[class_al]

    def determin_class_usnrc_fetter(self, rwi):
        """Determin the radwaste class USNRC, according to the rw.
        
        Parameters:
        -----------
        rwi: numpy array (1D)
            Indices of difference class of a specific cooling time.
        """

        classes = ['LLWA', 'LLWB', 'LLWC']
        # class according to rwi_ll
        for i in range(len(rwi)):
            if rwi[i] <= 1.0:
                return classes[i]

        # all rwi bigger than 1
        return 'ILW'


    def determin_class_uk(self, alpha_acts, acts, decay_heat):
        """Determin the radwaste class UK, according to the alpha activity,
        total specific activity and the decay heat.
        
        Parameters:
        -----------
        alpha_acts: float
            Specific acitvity of alpha decay nuclides.
        acts: float
            Specific activity of all the nuclides.
        decay_heat: float
            Decay heat of a cooling time. Unit: kW/m3.
        """

        # input check
        if alpha_acts < 0 or acts < 0 or decay_heat < 0:
            raise ValueError("Negtive data!")
        # if the decay heat is above 2kW/m3, it's HLW
        #if decay_heat > 2:
        #    return 'HLW'
        # if the decay heat is no more than 2kW/m3, check the alpha_acts and acts
        if alpha_acts <= 4.0e6 and (acts-alpha_acts) <= 1.2e7:
            return 'LLW'
        else:
            return 'ILW'

    def determin_class_russian(self, indices):
        """Determin the radwaste class RUSSIAN, according to the indices.
        
        Parameters:
        -----------
        indices: numpy array (1D)
            Indices of difference class of a specific cooling time.
            Eg. clear index, LLW index, ...
        """

        # check indices length
        if len(indices) != len(self.df.columns) -1:
            raise ValueError("indices length wrong")
        for i in range(len(indices)):
            if indices[i] <= 1.0:
                return self.df.columns[1+i]
        # not any index no more than 1, higher class
        return 'ILW'


def convert_nrc_fetter_limit(limit_cs, density):
    """
    Convert the NRC_FETTER limit from string to a float.
    """
    if limit_cs == 'TMSA':
        limit_c = float('inf')
        return limit_c
    
    tokens = limit_cs.strip().split()
    if len(tokens) == 1:
        value = float(limit_cs) # unit Ci/m3
        limit_c = ci2bq(value) / density / 1e3 # unit: Bq/kg
        return limit_c

    if len(tokens) == 2:
        value = float(tokens[0])
        unit = tokens[1] # unit is  (nCi/g)
        limit_c = 1e-9 * ci2bq(value) * 1e3 # unit: Bq/kg
        return limit_c

    # unknown limit_cs
    raise ValueError("limit_cs {0} format wrong".format(limit_cs))

def rwc_to_int(rwc, standard='CHN2018'):
    """Convert rwc to int.""" 
    if standard == 'CHN2018':
        rwc2inp = {'Clearance':1, 'VLLW':2, 'LLW':2, 'ILW':3, 'HLW':4}
    elif standard in ['USNRC', 'USNRC_FETTER']:
        rwc2inp = {'Clearance':1, 'LLWA':2, 'LLWB':2, 'LLWC':2, 'ILW':3}
    elif standard == 'RUSSIAN':
        rwc2inp = {'LLW':2, 'ILW':3}
    elif standard == 'UK':
        rwc2inp = {'LLW':2, 'ILW':3}
    else:
        raise ValueError(f"standard: {standard} not supported")
    return rwc2inp[rwc]

def ctr_to_int(ctr):
    """Convert cooling time requirement to int.""" 
    # convert ctr to float
    if '<' in ctr:
        ctr = float(ctr[1:]) * 0.99
    elif '>' in ctr:
        ctr = float(ctr[1:]) * 1.01
    else:
        ctr = float(ctr)

    # check ctr
    if ctr <= 1: # green
        return 1
    elif 1 < ctr and ctr <= 10: # blue
        return 2
    elif 10 < ctr and ctr <= 100: # yellow
        return 3
    elif 100 < ctr and ctr <= 1000: # pink
        return 4
    elif 1000 < ctr: # red
        return 5
    else:
        raise ValueError(f"Wrong ctr: {ctr}")

def is_u(nuc):
    """Check whether a nuclide is U"""
    u_pattern = re.compile("^U[0-9]*", re.IGNORECASE)
    if re.match(u_pattern, nuc):
        return True
    else:
        return False

def is_alpha_5(nuc):
    """Check whether a nuclide is alpha emitting and half-life > 5 year"""
    alpha_5_nucs = ('Np237', 'Pu238', 'Pu239', 'Pu240', 'Pu241', 'Pu242', 'Pu244',
                    'Am241', 'Am242m', 'Am243', 'Cm243', 'Cm244', 'Cm245', 'Cm246',
                    'Cm247', 'Cm248', 'Cf249', 'Cf250', 'Cf251')
    if nuc in alpha_5_nucs:
        return True
    else:
        return False



from setuptools import setup, find_packages
 
setup(name='natf',
      version='1.6.5',
      url='https://github.com/zxkjack123/NATF',
      license='MIT',
      author='Xiaokang Zhang',
      author_email='zxkjack123@163.com',
      description='Nuclear Analysis toolkit for Fusion with coupling of MCNP and FISPACT',
      packages=find_packages(exclude=['tests']),
      package_dir={'natf':'natf'},
      include_package_data=True,
      package_data={'natf':[
                        'radwaste_standards/CHN2018/*',
                        'radwaste_standards/USNRC/*',
                        'radwaste_standards/USNRC_FETTER/*',
                        'radwaste_standards/RUSSIAN/*',
                        'data/*',
                        ],
                    },
      long_description=open('README.md').read(),
      entry_points={
          "console_scripts":["natf_run = natf.natf_functions:natf_run",
                             "nonvoid_cells_to_tally = natf.mcnp_input:nonvoid_cells_to_tally",
                             "cell_vol_to_tally = natf.mcnp_input:cell_vol_to_tally",
                             "part_cell_list = natf.mcnp_input:part_cell_list",]
          },
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Environment :: Console",
          ],
      python_requires='>=3.6',
      zip_safe=False)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
from natf.material import Material
import numpy as np
from natf.utils import is_short_live, mcnp_style_str_append
from natf.half_space import HalfSpace


class Cell(object):
    ''' class Cell'''

    def __init__(self):
        self._name = ''
        self._icl = None
        self._id = None
        self._mat_id = None
        self._mat = Material()
        self._geom = [] # list of dict of {Boolean:surf_list}
        self._imp_n = 1.0
        self._imp_p = None
        self._vol = 0.0 # cm3
        self._mass = 0.0 # g
        self._density = 0.0 # g/cm3
        self._neutron_flux = np.zeros(0, dtype=float)
        self._neutron_flux_error = np.zeros(0, dtype=float)
        self._gamma_emit_rate = np.zeros((0, 0), dtype=float, order='C')
        self._nuclide = []  # fispact output data nuclide
        self._half_life = []
        # activity part, unit: Bq/kg
        # specific activity of nuclides at different inteval, shape=(INTV, NUC)
        self._act = np.zeros((0, 0), dtype=float, order='C')
        self._act_max_contri_nuc = []  # read only
        # specific activity of max contribution nuclide
        self._act_max_contri_act = np.zeros((0, 0), dtype=float,
                                            order='C')
        self._act_max_contri_ratio = np.zeros((0, 0), dtype=float)
        self._total_act = np.zeros(0, dtype=float)
        self._total_alpha_act = np.zeros((0), dtype=float)
        # decay heat part, unit: kW/kg
        self._decay_heat = np.zeros((0, 0), dtype=float)
        self._decay_heat_max_contri_nuc = []
        self._decay_heat_max_contri_dh = np.zeros((0, 0), dtype=float)
        self._decay_heat_max_contri_ratio = np.zeros((0, 0), dtype=float)
        self._total_decay_heat = np.zeros(0, dtype=float)
        # contact dose part,
        self._contact_dose = np.zeros((0, 0), dtype=float)
        self._contact_dose_max_contri_nuc = []
        self._contact_dose_max_contri_cd = np.zeros((0, 0), dtype=float)
        self._contact_dose_max_contri_ratio = np.zeros((0, 0), dtype=float)
        self._total_contact_dose = []
        # Clear Index part
        self._ci = np.zeros((0, 0), dtype=float)
        self._ci_max_contri_nuc = []
        self._ci_max_contri_ci = np.zeros((0, 0), dtype=float)
        self._ci_max_contri_ratio = np.zeros((0, 0), dtype=float)
        self._total_ci = np.zeros(0, dtype=float)
        # irradiation damage part
        self._dpa = None
        self._He_production = None
        self._H_production = None
        # radwaste part
        # CHN2018
        self._rw_chn2018_index_sum = None
        self._radwaste_class_chn2018 = []
        # USNRC
        self._rw_usnrc_index_sum_ll = None
        self._rw_usnrc_index_sum_sl = None
        self._radwaste_class_usnrc = []
        # USNRC_FETTER
        self._rw_usnrc_fetter_index_sum = None
        self._radwaste_class_usnrc_fetter = []
        # UK
        self._radwaste_class_uk = []
        # RUSSIAN
        self._rw_russian_index_sum = None
        self._radwaste_class_russian = []

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise ValueError('name must be string')
        self._name = value

    # icl getter and setter
    @property
    def icl(self):
        return self._icl

    @icl.setter
    def icl(self, value):
        if not isinstance(value, int):
            raise ValueError('icl must be integer')
        if value < 1 or value > 100000:
            raise ValueError('icl must between 1 and 100000')
        self._icl = value

    # id setter and getter
    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if not isinstance(value, int):
            raise ValueError('id must be integer')
        if value < 1 or value > 100000:
            raise ValueError('id must between 1 and 100000')
        self._id = value

    # mat_id setter and getter
    @property
    def mat_id(self):
        return self._mat_id

    @mat_id.setter
    def mat_id(self, value):
        if not isinstance(value, int):
            raise ValueError('mat_id must be integer')
        if value < 0 or value > 100000:
            raise ValueError('mat_id must between 0 and 100000')
        self._mat_id = value

    @property
    def mat(self):
        return self._mat

    @mat.setter
    def mat(self, value):
        if not isinstance(value, Material):
            raise ValueError('mat must be a object of class Material')
        self._mat = value

    @property
    def surf_list(self):
        return self._surf_list
    @surf_list.setter
    def self_list(self, value):
        if not isinstance(value, list):
            raise ValueError('surf_list of cell must be a list')
        for i, surf in value:
            if not isinstance(value, int):
                raise ValueError('surf_list should be a list of int')
        self._surf_list = value
 
    @property
    def surf_sign(self):
        return self._surf_sign
    @surf_sign.setter
    def self_sign(self, value):
        if not isinstance(value, list):
            raise ValueError('surf_sign of cell must be a list')
        for i, surf in value:
            if not isinstance(value, str):
                raise ValueError('surf_sign should be a list of string')
            if value not in ('*', '+', ''):
                raise ValueError('surf sign {0} not support'.format(value))
        self._surf_sign = value

    @property
    def imp_n(self):
        return self._imp_n

    @imp_n.setter
    def imp_n(self, value):
        if value < 0:
            raise ValueError('imp_n must be a number that no smaller than 0')
        self._imp_n = value

    @property
    def imp_p(self):
        return self._imp_p

    @imp_p.setter
    def imp_p(self, value):
        if value < 0:
            raise ValueError('imp_p must be a number that no smaller than 0')
        self._imp_p = value


    @property
    def vol(self):
        return self._vol

    @vol.setter
    def vol(self, value):
        if not isinstance(value, float):
            raise ValueError('vol of cell must be a float')
        if value < 0:
            raise ValueError('vol must be a float that no smaller than 0')
        self._vol = value

    @property
    def mass(self):
        return self._mass

    @mass.setter
    def mass(self, value):
        if not isinstance(value, float):
            raise ValueError('mass of cell must be a float')
        if value < 0:
            raise ValueError('mass of cell must be a float no smaller than 0')
        self._mass = value

    @property
    def density(self):
        return self._density

    @density.setter
    def density(self, value):
        if not isinstance(value, float):
            raise ValueError('density of cell must be a float')
        if value < 0 or value > 30:
            raise ValueError('density of cell must between 0 and 30')
        self._density = value

    @property
    def neutron_flux(self):
        return self._neutron_flux

    @neutron_flux.setter
    def neutron_flux(self, value):
        if not isinstance(value, list):
            raise ValueError('nuetron_flux of cell must be a list')
        if len(value) not in [176, 316, 710]:
            raise ValueError('neutron flux should have data of 176/316/710, \
                    with the last of total data')
        for i in range(len(value)):
            if not isinstance(value[i], float):
                raise ValueError(
                    'neutron_flux of cell must be a list of float')
            if value[i] < 0:
                raise ValueError('neutron_flux must be a list that each element\
                        is no smaller than 0.0')
        self._neutron_flux = value

    @property
    def neutron_flux_error(self):
        return self._neutron_flux_error

    @neutron_flux_error.setter
    def neutron_flux_error(self, value):
        if not isinstance(value, list):
            raise ValueError('neutron_flux_error must be a list')
        if len(value) not in [176, 316, 710]:
            raise ValueError('neutron flux should have data of 176/316/710, \
                    the last of total data')
        for i in range(len(value)):
            if not isinstance(value[i], float):
                raise ValueError('neutron_flux_error must be a list of float')
            if value[i] < 0.0 or value[i] > 1:
                raise ValueError('neutron_flux_error must between 0 and 1')
        self._neutron_flux_error = value

    @property
    def gamma_emit_rate(self):
        return self._gamma_emit_rate

    @gamma_emit_rate.setter
    def gamma_emit_rate(self, value):
        if not isinstance(value, type(np.ndarray((0, 0)))):
            raise ValueError('gamma_emit_rate must be ndarray')
        if len(value.shape) != 2:
            raise ValueError('gamma_emit_rate must be a two-dimensional\
                    ndarray')
        for i in range(value.shape[0]):
            for j in range(value.shape[1]):
                if not isinstance(value[i, j], float):
                    raise ValueError('gamma_emit_rate must composed of float')
        if value.min() < 0.0:
            raise ValueError('gamma_emit_rate must no smaller than 0.0')
        self._gamma_emit_rate = value

    @property
    def nuclide(self):
        return self._nuclide

    @nuclide.setter
    def nuclide(self, value):
        if not isinstance(value, list):
            raise ValueError('nuclide of cell must be a list')
        for i in range(len(value)):
            if not isinstance(value[i], str):
                raise ValueError('nuclide of cell should be a list of string')
        self._nuclide = value

    @property
    def half_life(self):
        return self._half_life

    @half_life.setter
    def half_life(self, value):
        if not isinstance(value, list):
            raise ValueError('half_life must a list')
        for i in range(len(value)):
            if not isinstance(value[i], float):
                raise ValueError('half_life of nuclide must be float')
            if value[i] < 0:
                raise ValueError('half_life of nuclide shoule be no smaller\
                        than 0.0')
        self._half_life = value

    @property
    def act(self):
        return self._act

    @act.setter
    def act(self, value):
        if not isinstance(value, type(np.ndarray((0, 0)))):
            raise ValueError('act must be ndarray')
        if len(value.shape) != 2:
            raise ValueError('act must be a two-dimensional ndarray')
        for i in range(value.shape[0]):
            for j in range(value.shape[1]):
                if not isinstance(value[i, j], float):
                    raise ValueError('act must composed of float')
        if value.min() < 0.0:
            raise ValueError('act must no smaller than 0.0')
        self._act = value

    @property
    def act_max_contri_nuc(self):
        return self._act_max_contri_nuc

    @property
    def act_max_contri_act(self):
        return self._act_max_contri_act

    @property
    def act_max_contri_ratio(self):
        return self._act_max_contri_ratio

    @property
    def total_act(self):
        return self._total_act

    @property
    def total_alpha_act(self):
        return self._total_alpha_act

    @total_alpha_act.setter
    def total_alpha_act(self, value):
        if not isinstance(value, type(np.ndarray((0)))):
            raise ValueError('total_alpha_act must be ndarray')
        if len(value.shape) != 1:
            raise ValueError('total_alpha_act must be a one-dimensional ndarray')
        for i in range(value.shape[0]):
            if not isinstance(value[i], float):
                raise ValueError('total_alpha_act must composed of float')
        if value.min() < 0.0:
            raise ValueError('total_alpha_act must no smaller than 0.0')
        self._total_alpha_act = value


    @property
    def decay_heat(self):
        return self._decay_heat

    @decay_heat.setter
    def decay_heat(self, value):
        if not isinstance(value, type(np.ndarray((0, 0)))):
            raise ValueError('act must be ndarray')
        if len(value.shape) != 2:
            raise ValueError('act must be a two-dimensional ndarray')
        for i in range(value.shape[0]):
            for j in range(value.shape[1]):
                if not isinstance(value[i, j], float):
                    raise ValueError('act must composed of float')
        if value.min() < 0.0:
            raise ValueError('act must no smaller than 0.0')
        self._decay_heat = value

    @property
    def decay_heat_max_contri_nuc(self):
        return self._decay_heat_max_contri_nuc

    @property
    def decay_heat_max_contri_dh(self):
        return self._decay_heat_max_contri_dh

    @property
    def decay_heat_max_contri_ratio(self):
        return self._decay_heat_max_contri_ratio

    @property
    def total_decay_heat(self):
        return self._total_decay_heat


    @property
    def contact_dose(self):
        return self._contact_dose

    @contact_dose.setter
    def contact_dose(self, value):
        if not isinstance(value, type(np.ndarray((0, 0)))):
            raise ValueError('contact_dose must be ndarray')
        if len(value.shape) != 2:
            raise ValueError('contact_dose must be a two-dimensional ndarray')
        for i in range(value.shape[0]):
            for j in range(value.shape[1]):
                if not isinstance(value[i, j], float):
                    raise ValueError('contact_dose must composed of float')
        if value.min() < 0.0:
            raise ValueError('contact_dose must no smaller than 0.0')
        self._contact_dose = value

    @property
    def contact_dose_max_contri_nuc(self):
        return self._contact_dose_max_contri_nuc

    @property
    def contact_dose_max_contri_cd(self):
        return self._contact_dose_max_contri_cd

    @property
    def contact_dose_max_contri_ratio(self):
        return self._contact_dose_max_contri_ratio

    @property
    def total_contact_dose(self):
        return self._total_contact_dose

    @property
    def ci(self):
        return self._ci

    @ci.setter
    def ci(self, value):
        if not isinstance(value, type(np.ndarray((0, 0)))):
            raise ValueError('contact_dose must be ndarray')
        if len(value.shape) != 2:
            raise ValueError('contact_dose must be a two-dimensional ndarray')
        for i in range(value.shape[0]):
            for j in range(value.shape[1]):
                if not isinstance(value[i, j], float):
                    raise ValueError('contact_dose must composed of float')
        if value.min() < 0.0:
            raise ValueError('contact_dose must no smaller than 0.0')
        self._ci = value

    @property
    def ci_max_contri_nuc(self):
        return self._ci_max_contri_nuc

    @property
    def ci_max_contri_ci(self):
        return self._ci_max_contri_ci

    @property
    def ci_max_contri_ratio(self):
        return self._ci_max_contri_ratio

    @property
    def total_ci(self):
        return self._total_ci

    @property
    def dpa(self):
        return self._dpa

    @dpa.setter
    def dpa(self, value):
        if not isinstance(value, float):
            raise ValueError('dpa value must be a float number')
        if value < 0:
            raise ValueError('dpa value must no smaller than 0.0')
        self._dpa = value

    @property
    def He_production(self):
        return self._He_production

    @He_production.setter
    def He_production(self, value):
        if not isinstance(value, float):
            raise ValueError('He_production value must be a float number')
        if value < 0:
            raise ValueError('He_production value must no smaller than 0.0')
        self._He_production = value

    @property
    def H_production(self):
        return self._H_production

    @H_production.setter
    def H_production(self, value):
        if not isinstance(value, float):
            raise ValueError('H_production value must be a float number')
        if value < 0:
            raise ValueError('H_production value must no smaller than 0.0')
        self._H_production = value

    def analysis_act(self):
        """get the act_max_contri_nuc, act_max_contri_act and
        act_max_contri_ratio"""
        # input check for nuclide, a
        if self._act.size == 0:
            raise ValueError('act must assignment before calculate the\
                    total_act')
        self._total_act = self._act.sum(axis=1)  # calculate the a_total
        # get the max nuclide
        for i in range(self._act.shape[0]):
            nid = np.where(self._act[i] == np.max(self._act[i]))
            nid = _pick_max_ind_from_tuple(nid)
            # judge whether this nuclide exis in the list
            if self._nuclide[nid] not in self._act_max_contri_nuc:
                self._act_max_contri_nuc.append(self._nuclide[nid])
                # cal the act_max_contri_act
                if i == 0:
                    self._act_max_contri_act = self._act[:, nid]
                    self._act_max_contri_act = np.column_stack(
                        (self._act_max_contri_act, self._act[:, nid]))
                    self._act_max_contri_act = np.delete(
                        self._act_max_contri_act, 1, axis=1)
                else:
                    self._act_max_contri_act = np.column_stack((
                        self._act_max_contri_act, self._act[:, nid]))
        # cal the act_max_contri_ratio
        self._act_max_contri_ratio = self._act_max_contri_act.copy()
        for i in range(self._act_max_contri_act.shape[0]):
            for j in range(self._act_max_contri_act.shape[1]):
                self._act_max_contri_ratio[i][j] = \
                    self._act_max_contri_act[i][j] / self._total_act[i]

    @property
    def rw_chn2018_index_sum(self):
        return self._rw_chn2018_index_sum

    @rw_chn2018_index_sum.setter
    def rw_chn2018_index_sum(self, value):
        if not isinstance(value, type(np.ndarray((0, 0)))):
            raise ValueError('rw_chn2018_index_sum must be ndarray')
        if len(value.shape) != 2:
            raise ValueError('rw_chn2018_index_sum must be a two-dimensional ndarray')
        for i in range(value.shape[0]):
            for j in range(value.shape[1]):
                if not isinstance(value[i, j], float):
                    raise ValueError('rw_chn2018_index_sum must composed of float')
        if value.min() < 0.0:
            raise ValueError('rw_chn2018_index_sum must no smaller than 0.0')
        self._rw_chn2018_index_sum = value

    @property
    def rw_usnrc_index_sum_ll(self):
        return self._rw_usnrc_index_sum_ll

    @rw_usnrc_index_sum_ll.setter
    def rw_usnrc_index_sum_ll(self, value):
        if not isinstance(value, type(np.ndarray((0, 0)))):
            raise ValueError('rw_usnrc_index_sum_ll must be ndarray')
        if len(value.shape) != 2:
            raise ValueError('rw_usnrc_index_sum_ll must be a two-dimensional ndarray')
        for i in range(value.shape[0]):
            for j in range(value.shape[1]):
                if not isinstance(value[i, j], float):
                    raise ValueError('rw_usnrc_index_sum_ll must composed of float')
        if value.min() < 0.0:
            raise ValueError('rw_usnrc_index_sum_ll must no smaller than 0.0')
        self._rw_usnrc_index_sum_ll = value

    @property
    def rw_usnrc_index_sum_sl(self):
        return self._rw_usnrc_index_sum_sl

    @rw_usnrc_index_sum_sl.setter
    def rw_usnrc_index_sum_sl(self, value):
        if not isinstance(value, type(np.ndarray((0, 0)))):
            raise ValueError('rw_usnrc_index_sum_sl must be ndarray')
        if len(value.shape) != 2:
            raise ValueError('rw_usnrc_index_sum_sl must be a two-dimensional ndarray')
        for i in range(value.shape[0]):
            for j in range(value.shape[1]):
                if not isinstance(value[i, j], float):
                    raise ValueError('rw_usnrc_index_sum_sl must composed of float')
        if value.min() < 0.0:
            raise ValueError('rw_usnrc_index_sum_sl must no smaller than 0.0')
        self._rw_usnrc_index_sum_sl = value

    @property
    def rw_usnrc_fetter_index_sum(self):
        return self._rw_usnrc_fetter_index_sum

    @rw_usnrc_fetter_index_sum.setter
    def rw_usnrc_fetter_index_sum(self, value):
        if not isinstance(value, type(np.ndarray((0, 0)))):
            raise ValueError('rw_usnrc_fetter_index_sum must be ndarray')
        if len(value.shape) != 2:
            raise ValueError('rw_usnrc_fetter_index_sum must be a two-dimensional ndarray')
        for i in range(value.shape[0]):
            for j in range(value.shape[1]):
                if not isinstance(value[i, j], float):
                    raise ValueError('rw_usnrc_fetter_index_sum must composed of float')
        if value.min() < 0.0:
            raise ValueError('rw_usnrc_fetter_index_sum must no smaller than 0.0')
        self._rw_usnrc_fetter_index_sum = value

    @property
    def rw_russian_index_sum(self):
        return self._rw_russian_index_sum

    @rw_russian_index_sum.setter
    def rw_russian_index_sum(self, value):
        if not isinstance(value, type(np.ndarray((0, 0)))):
            raise ValueError('rw_russian_index_sum must be ndarray')
        if len(value.shape) != 2:
            raise ValueError('rw_russian_index_sum must be a two-dimensional ndarray')
        for i in range(value.shape[0]):
            for j in range(value.shape[1]):
                if not isinstance(value[i, j], float):
                    raise ValueError('rw_russian_index_sum must composed of float')
        if value.min() < 0.0:
            raise ValueError('rw_russian_index_sum must no smaller than 0.0')
        self._rw_russian_index_sum = value


    def analysis_decay_heat(self):
        # input check for nuclide,
        if self._decay_heat.size == 0:
            raise ValueError('decay_heat must assignment before calculate the\
                    total_act')
        # calculate the a_total
        self._total_decay_heat = self._decay_heat.sum(axis=1)
        # get the max nuclide
        for i in range(self._decay_heat.shape[0]):
            nid = np.where(self._decay_heat[i] == np.max(self._decay_heat[i]))
            nid = _pick_max_ind_from_tuple(nid)
            # judge whether this nuclide exis in the list
            if self._nuclide[nid] not in self._decay_heat_max_contri_nuc:
                self._decay_heat_max_contri_nuc.append(self._nuclide[nid])
                # cal the act_max_contri_hd
                if i == 0:
                    self._decay_heat_max_contri_dh = self._decay_heat[:, nid]
                    self._decay_heat_max_contri_dh = np.column_stack(
                        (self._decay_heat_max_contri_dh,
                            self._decay_heat[:, nid]))
                    self._decay_heat_max_contri_dh = np.delete(
                        self._decay_heat_max_contri_dh, 1, axis=1)
                else:
                    self._decay_heat_max_contri_dh = np.column_stack(
                        (self._decay_heat_max_contri_dh,
                            self._decay_heat[:, nid]))
        # cal the act_max_contri_ratio
        self._decay_heat_max_contri_ratio = \
            self._decay_heat_max_contri_dh.copy()
        for i in range(self._decay_heat_max_contri_dh.shape[0]):
            for j in range(self._decay_heat_max_contri_dh.shape[1]):
                self._decay_heat_max_contri_ratio[i][j] = \
                    self._decay_heat_max_contri_dh[i][j] / \
                    self._total_decay_heat[i]

    def analysis_contact_dose(self):
        # input check for nuclide,
        if self._contact_dose.size == 0:
            raise ValueError('contact_dose must assignment before calculate \
                    the total_act')
        # calculate the a_total
        self._total_contact_dose = self._contact_dose.sum(axis=1)
        # get the max nuclide
        for i in range(self._contact_dose.shape[0]):
            nid = np.where(self._contact_dose[i] == np.max(
                self._contact_dose[i]))
            nid = _pick_max_ind_from_tuple(nid)
            # judge whether this nuclide exis in the list
            if self._nuclide[nid] not in self._contact_dose_max_contri_nuc:
                self._contact_dose_max_contri_nuc.append(self._nuclide[nid])
                # cal the act_max_contri_hd
                if i == 0:
                    self._contact_dose_max_contri_cd = \
                        self._contact_dose[:, nid]
                    self._contact_dose_max_contri_cd = np.column_stack(
                        (self._contact_dose_max_contri_cd,
                         self._contact_dose[:, nid]))
                    self._contact_dose_max_contri_cd = \
                        np.delete(self._contact_dose_max_contri_cd, 1,
                                  axis=1)
                else:
                    self._contact_dose_max_contri_cd = np.column_stack(
                        (self._contact_dose_max_contri_cd,
                         self._contact_dose[:, nid]))
        # cal the act_max_contri_ratio
        self._contact_dose_max_contri_ratio = \
            self._contact_dose_max_contri_cd.copy()
        for i in range(self._contact_dose_max_contri_cd.shape[0]):
            for j in range(self._contact_dose_max_contri_cd.shape[1]):
                self._contact_dose_max_contri_ratio[i][j] = \
                    self._contact_dose_max_contri_cd[i][j] / self._total_contact_dose[i]

    def analysis_ci(self):
        # input check for nuclide, a
        if self._ci.size == 0:
            raise ValueError(
                'ci must assignment before calculate the total_ci')
        self._total_ci = self._ci.sum(axis=1)  # calculate the a_total
        # get the max nuclide
        for i in range(self._ci.shape[0]):
            nid = np.where(self._ci[i] == np.max(self._ci[i]))
            nid = _pick_max_ind_from_tuple(nid)
            # judge whether this nuclide exis in the list
            if self._nuclide[nid] not in self._ci_max_contri_nuc:
                self._ci_max_contri_nuc.append(self._nuclide[nid])
                # cal the ci_max_contri_ci
                if i == 0:
                    self._ci_max_contri_ci = self._ci[:, nid]
                    self._ci_max_contri_ci = np.column_stack(
                        (self._ci_max_contri_ci, self._ci[:, nid]))
                    self._ci_max_contri_ci = np.delete(self._ci_max_contri_ci,
                                                       1, axis=1)
                else:
                    self._ci_max_contri_ci = np.column_stack((
                        self._ci_max_contri_ci, self._ci[:, nid]))
        # cal the ci_max_contri_ratio
        self._ci_max_contri_ratio = self._ci_max_contri_ci.copy()
        for i in range(self._ci_max_contri_ci.shape[0]):
            for j in range(self._ci_max_contri_ci.shape[1]):
                self._ci_max_contri_ratio[i][j] = \
                    self._ci_max_contri_ci[i][j] / self._total_ci[i]

    def analysis_radwaste(self, rws=None):
        """
        Analysis the radwaste classification.

        Parameters:
        -----------
        rws: RadwasteStandard
            The standard used.
            Supported standards are: CHN2018, UK, USNRC, USNRC_FETTER, RUSSIAN.
        """
        if rws.standard == 'CHN2018':
            self.analysis_radwaste_chn2018(rws)
        elif rws.standard == 'USNRC':
            self.analysis_radwaste_usnrc(rws)
        elif rws.standard == 'USNRC_FETTER':
            self.analysis_radwaste_usnrc_fetter(rws)
        elif rws.standard == 'UK':
            self.analysis_radwaste_uk(rws)
        elif rws.standard == 'RUSSIAN':
            self.analysis_radwaste_russian(rws)
        else:
            raise ValueError("rws {0} not supported!".format(rws.standard))

    def analysis_radwaste_chn2018(self, rws=None):
        """
        Analysis the radwaste classification.

        Parameters:
        -----------
        rws: RadwasteStandard
            The standard used, must be CHN2018.
        """

        # init rw_index with shape of (INTV, NUC, class)
        rw_index = np.zeros(shape=(len(self.act),
            len(self.act[0]), len(rws.classes)), dtype=float)

        # calculate radwaste index for each nuclide
        for nid, nuc in enumerate(self.nuclide):
            # get the limits for the nuclide
            limits = rws.get_nuc_limits(nuc=nuc)
            # loop over the interval
            for intv in range(len(self.act)):
                # calculate the index for each class
                rw_index[intv, nid, :] = np.divide(self.act[intv, nid], limits)

        # sum up the index for each nuclide, shape=(INTV, class)
        rw_index_sum = np.sum(rw_index, axis=1)

        # get the radwaste classification according to the indices for each class
        rw_class = []
        for intv in range(len(self.act)):
            # CHN2018 use rw_index_sum and decay heat to classify radwaste
            # convert decay heat from kW/kg to kW/m3
            decay_heat = self.total_decay_heat[intv] # kW/kg
            decay_heat = decay_heat * self.density * 1000.0 # kW/m3
            rw_class.append(rws.determin_class_chn2018(rw_index_sum[intv],
                decay_heat))

        self.rw_chn2018_index_sum = rw_index_sum.copy()
        self.radwaste_class_chn2018 = rw_class

    def analysis_radwaste_usnrc(self, rws=None):
        """
        Analysis the radwaste classification.

        Parameters:
        -----------
        rws: RadwasteStandard
            Supported standards is: USNRC.
        """

        # init rw_index with shape of (INTV, NUC, class)
        rw_index_sl = np.zeros(shape=(len(self.act),
            len(self.act[0]), len(rws.classes)), dtype=float)
        rw_index_ll = np.zeros(shape=(len(self.act),
            len(self.act[0]), len(rws.classes)), dtype=float)

        # calculate radwaste index for each nuclide
        for nid, nuc in enumerate(self.nuclide):
            # get the limits for the nuclide
            limits = rws.get_nuc_limits(nuc=nuc, half_life=self.half_life[nid],
                    density=self.density)
            # loop over the interval
            for intv in range(len(self.act)):
                # calculate the index for each class
                if nuc in ('Pu241', 'Cm242') or (not is_short_live(self.half_life[nid])):
                    rw_index_ll[intv, nid, :] = np.divide(self.act[intv, nid], limits)
                else:
                    rw_index_sl[intv, nid, :] = np.divide(self.act[intv, nid], limits)

        rw_usnrc_index_sum_ll = np.sum(rw_index_ll, axis=1)
        rw_usnrc_index_sum_sl = np.sum(rw_index_sl, axis=1)

        # get the radwaste classification according to the indices for each class
        rw_class = []
        for intv in range(len(self.act)):
            # USNRC use rw_usnrc_index_sum_sl and rw_usnrc_index_sum_ll to calssify radwaste
            rw_class.append(rws.determin_class_usnrc(rw_usnrc_index_sum_ll[intv],
                rw_usnrc_index_sum_sl[intv]))

        self.rw_usnrc_index_sum_ll = rw_usnrc_index_sum_ll.copy()
        self.rw_usnrc_index_sum_sl = rw_usnrc_index_sum_sl.copy()
        self.radwaste_class_usnrc = rw_class
    
    def analysis_radwaste_usnrc_fetter(self, rws=None):
        """
        Analysis the radwaste classification.

        Parameters:
        -----------
        rws: RadwasteStandard
            Supported standards is: USNRC_FETTER.
        """

        # init rw_index with shape of (INTV, NUC, class)
        rw_index = np.zeros(shape=(len(self.act),
            len(self.act[0]), len(rws.classes)), dtype=float)

        # calculate radwaste index for each nuclide
        for nid, nuc in enumerate(self.nuclide):
            # get the limits for the nuclide
            limits = rws.get_nuc_limits(nuc=nuc, half_life=self.half_life[nid],
                    density=self.density)
            # loop over the interval
            for intv in range(len(self.act)):
                # calculate the index for each class
                rw_index[intv, nid, :] = np.divide(self.act[intv, nid], limits)

        rw_index_sum = np.sum(rw_index, axis=1)

        # get the radwaste classification according to the indices for each class
        rw_class = []
        for intv in range(len(self.act)):
            # USNRC use rw_usnrc_index_sum_sl and rw_usnrc_index_sum_ll to calssify radwaste
            rw_class.append(rws.determin_class_usnrc_fetter(rw_index_sum[intv]))

        self.rw_usnrc_fetter_index_sum = rw_index_sum.copy()
        self.radwaste_class_usnrc_fetter = rw_class
 
    def analysis_radwaste_uk(self, rws=None):
        """
        Analysis the radwaste classification.

        Parameters:
        -----------
        rws: RadwasteStandard
            The standard used, must be UK.
        """

        # get the radwaste classification according to the:
        # alpha activity, activity and decay heat
        rw_class = []
        for intv in range(len(self.total_act)):
            # convert decay heat from kW/kg to kW/m3
            decay_heat = self.total_decay_heat[intv] # kW/kg
            decay_heat = decay_heat * self.density * 1000.0 # kW/m3
            rw_class.append(rws.determin_class_uk(self.total_alpha_act[intv],
                self.total_act[intv], decay_heat))
        self.radwaste_class_uk = rw_class

    def analysis_radwaste_russian(self, rws=None):
        """
        Analysis the radwaste classification.

        Parameters:
        -----------
        rws: RadwasteStandard
            The standard used, must be RUSSIAN.
        """

        # init rw_index with shape of (INTV, NUC, class)
        rw_index = np.zeros(shape=(len(self.act),
            len(self.act[0]), 1), dtype=float)

        # calculate radwaste index for each nuclide
        for nid, nuc in enumerate(self.nuclide):
            # get the limits for the nuclide
            limits = rws.get_nuc_limits(nuc=nuc)
            # loop over the interval
            for intv in range(len(self.act)):
                # calculate the index for each class
                rw_index[intv, nid, :] = np.divide(self.act[intv, nid], limits)

        # sum up the index for each nuclide, shape=(INTV, class)
        rw_index_sum = np.sum(rw_index, axis=1)

        # get the radwaste classification according to the indices for each class
        rw_class = []
        for intv in range(len(self.act)):
            rw_class.append(rws.determin_class_russian(rw_index_sum[intv]))
        self.rw_russian_index_sum = rw_index_sum.copy()
        self.radwaste_class_russian = rw_class


    def __str__(self):
        """print mcnp style cell card"""
        s = ''.join([str(self.id), '     ', str(self.mat.id), ' ', '-', str(self.mat.density)])
        indent_length = len(s)
        for key, value in self.geom:
            if key == 'intersection': 
                bool_mark = ''
            elif key == 'union':
                bool_mark = ':'
            elif key == 'complement':
                bool_mark = '#'
            for j, hs in enumerate(value):
                if hs.sense == '-':
                    hs_sense_str = hs.sense
                else:
                    hs_sense_str = ''
                hs_str = ''.join([bool_mark, hs_sense_str, str(hs.surf.id)])
                s = mcnp_style_str_append(s, hs_str, indent_length)
        s = mcnp_style_str_append(s, ''.join(['imp:n=', str(self.imp_n)]))
        if self.imp_p is not None:
            s = mcnp_style_str_append(s, ''.join(['imp:p=', str(self.imp_p)]))
        return s    


def _pick_max_ind_from_tuple(nid):
    '''
    Pick up the index of max row.
    The nid is a tuple returned from np.where
    such as: ([1]), ([1, 2], )
    '''
    if len(nid[0]) == 1:
        nid = int(nid[0])
    else:
        if len(nid[0]) > 1:  # in case that two nuclide have same max act
            nid = int(nid[0][0])
    return nid

def get_cell_index(Cells, cid):
    """get_cell_index, get the cell index in the Cells according to the id"""
    cell_index = -1
    for i in range(len(Cells)):
        if Cells[i].id == cid:
            cell_index = i
            break
    # check value
    if cell_index == -1:  # this means no cell found
        raise ValueError('cell not found!')
    return cell_index


